from typing import Any

from ...base import BaseConversationLengthAdjuster
from ..model_config import MODEL_CONFIG, MODEL_POINT
from ..openai_utils import get_encoding, get_n_tokens


class OldConversationTruncationModule(BaseConversationLengthAdjuster):
    """
    Adjust the number of tokens to be less than or equal to context_length, starting from the oldest message forward

    FIXME: Truncation is not accurate especially when tool is called.
    """

    def __init__(self, model_name: str, context_length: int):
        if model_name in MODEL_POINT.keys():
            print(f"{model_name} is automatically converted to {MODEL_POINT[model_name]}")
            model_name = MODEL_POINT[model_name]

        assert (
            model_name in MODEL_CONFIG.keys()
        ), f"model_name must be one of {', '.join(sorted(MODEL_CONFIG.keys()))}."

        self.model_name = model_name
        self.context_length = context_length
        self.encoding = get_encoding(self.model_name)

    def run(self, messages: list[dict[str, dict[str, str]]]) -> list[dict[str, dict[str, str]]]:
        adjusted_messages: list[dict[str, dict[str, str]]] = []
        total_n_tokens: int = 0

        used_tool_call_ids = []
        used_tool_use_message = []
        total_tool_use_tokens = 0
        for message in messages[::-1]:
            if message.get("role") == "tool":
                n_tool_use_tokens = get_n_tokens(message, self.model_name)["total"]
                total_tool_use_tokens += n_tool_use_tokens
                if total_n_tokens + total_tool_use_tokens <= self.context_length:
                    used_tool_use_message.append(message)
                    continue
                else:
                    new_message = None
                    break
            elif "tool_calls" in message:
                for n_iter in range(1, len(message["tool_calls"]) + 1):
                    tool_calls = message["tool_calls"][-n_iter:]
                    _tmp_message = {k: v for k, v in message.items() if k != "tool_calls"}
                    _tmp_message["tool_calls"] = tool_calls

                    total_tool_call_tokens = get_n_tokens(_tmp_message, self.model_name)["total"]
                    if (
                        total_n_tokens + total_tool_use_tokens + total_tool_call_tokens
                        <= self.context_length
                    ):
                        used_tool_call_ids = [tool_call["id"] for tool_call in tool_calls]
                        tool_call_message = _tmp_message
                    else:
                        break

                if used_tool_call_ids:
                    new_message = [
                        tool_use_message
                        for tool_use_message in used_tool_use_message
                        if tool_use_message["tool_call_id"] in used_tool_call_ids
                    ]
                    new_message.append(tool_call_message)
                    for m in new_message:
                        n_tokens = get_n_tokens(m, self.model_name)["total"]
                        total_n_tokens += n_tokens

                else:
                    new_message = None
                    break
            else:
                new_message, total_n_tokens = self.adjust_message_length_and_update_total_tokens(
                    message, total_n_tokens
                )

            if new_message:
                if isinstance(new_message, dict):
                    adjusted_messages.append(new_message)
                elif isinstance(new_message, list):
                    adjusted_messages.extend(new_message)

                used_tool_call_ids = []
                used_tool_use_message = []
                total_tool_use_tokens = 0

            if new_message is None:
                break

        return adjusted_messages[::-1]

    def _truncate_text(self, text: str, available_n_tokens: int) -> tuple[dict[str, str], int]:
        new_text = self.truncate(text, available_n_tokens)
        print("Input message is truncated because total length of messages exceeds context length.")
        return new_text

    def _to_text_message(self, text) -> dict[str, str]:
        return {"type": "text", "text": text}

    def adjust_message_length_and_update_total_tokens(
        self, message: dict[str, dict[str, str]], total_n_tokens: int = 0
    ) -> dict[str, Any]:
        n_tokens = get_n_tokens(message, self.model_name)

        if total_n_tokens + n_tokens["total"] <= self.context_length:
            total_n_tokens += n_tokens["total"]
            return message, total_n_tokens
        else:
            available_n_tokens = max(
                self.context_length - total_n_tokens - n_tokens["other"], 0
            )  # available_n_tokens for content
            if available_n_tokens > 0:
                new_message = {}
                for key, value in message.items():
                    if key == "role":
                        new_message[key] = value
                    elif key == "name":
                        new_message[key] = value
                    elif key == "content" and isinstance(value, list):
                        new_contents = []
                        for item in value:
                            if item.get("type") == "image_url":
                                pass  # Image is entirely truncated
                            elif item.get("type") == "text":
                                total_n_tokens += available_n_tokens + n_tokens["other"]
                                new_text = self._truncate_text(
                                    text=item["text"],
                                    available_n_tokens=available_n_tokens,
                                )

                                new_contents.append({"type": "text", "text": new_text})
                            else:
                                raise ValueError(
                                    f"Unknown type {item['type']} in message['content']."
                                )
                        new_message[key] = new_contents

                    else:
                        raise ValueError(f"Unknown type {item['type']} in message['content'].")

                return new_message, total_n_tokens

            else:
                return None, total_n_tokens

    def truncate(self, text: str, n_tokens: int) -> str:
        if n_tokens > 0:
            return self.encoding.decode(self.encoding.encode(text)[-n_tokens:])
        else:
            return ""



import contextlib
import copy
import sqlite3
from pathlib import Path
from typing import Any

import numpy as np

from ...base import BaseMetadataStore


@contextlib.contextmanager
def sqlite_open(path: Path):
    conn = sqlite3.connect(path)
    cursor = conn.cursor()
    try:
        yield cursor
        conn.commit()
    finally:
        conn.rollback()
        conn.close()


class SQLiteMetadataStore(BaseMetadataStore):
    def __init__(self, persistence_directory: str, collection_name: str):
        self.persistence_directory = persistence_directory
        self.collection_name = collection_name
        self._path = Path(persistence_directory) / collection_name / "metadata.sqlite3"
        self._path.parent.mkdir(parents=True, exist_ok=True)
        self.p_key = "__id"

    def store(self, ids: list[int], metadatas: list[dict[str, str]]) -> None:
        _metadatas = copy.deepcopy(metadatas)
        all_keys = list(set([k for metadata in _metadatas for k in metadata.keys()]))

        # check if all keys are the same
        for _id, metadata in zip(ids, _metadatas, strict=True):
            for key in all_keys:
                if key not in metadata:
                    metadata[key] = None

            metadata[self.p_key] = _id

        # insert primary key
        all_keys = [self.p_key] + all_keys

        # create table
        db_columns = ",".join([key for key in all_keys])
        db_columns_with_property = ",".join(
            [key + " PRIMARY KEY" if key == self.p_key else key for key in all_keys]
        )

        with sqlite_open(self._path) as cursor:
            cursor.execute(f"CREATE TABLE IF NOT EXISTS metadata({db_columns_with_property})")
            cursor.executemany(
                f"INSERT OR REPLACE INTO metadata({db_columns}) VALUES ({','.join(['?']*len(all_keys))})",
                [tuple(metadata[key] for key in all_keys) for metadata in _metadatas],
            )

    def fetch(self, ids: list[int]) -> list[dict[str, Any]]:
        with sqlite_open(self._path) as cursor:
            cursor.execute(
                f"SELECT * FROM metadata WHERE {self.p_key} IN ({','.join(['?']*len(ids))})",
                tuple(ids),
            )
            _metadatas = cursor.fetchall()
            descriptions = cursor.description
            columns = [description[0] for description in descriptions]

        p_key_index = columns.index(self.p_key)

        _metadatas = [
            {
                columns[i]: metadata[i]
                for i in range(len(metadata))
                if i != p_key_index and metadata[i] is not None
            }
            for metadata in _metadatas
        ]

        _metadatas = [_metadatas[i] for i in np.argsort(np.argsort(ids))]
        return _metadatas



import copy
from typing import Any, AsyncGenerator, Generator, Iterable, Mapping, Union

import httpx
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import (
    NOT_GIVEN,
    NotGiven,
    ProxiesTypes,
    Timeout,
    Transport,
)
from anthropic.types import (
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
)
from anthropic.types.message_param import MessageParam
from anthropic.types.text_block_param import TextBlockParam

from ...base import (
    BaseChatModule,
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
    BaseMessage,
)
from ...llm_wrapper import ChatWrapperModule
from ...result import CompletionResults
from ...usage import TokenCounter, Usage
from ..claude_utils import acompletion, completion, get_async_client, get_client
from ..message import ClaudeMessage


class AnthropicChatCoreModule(BaseChatModule):
    def __init__(
        self,
        model_name: str,
        api_type: str = "anthropic",
        api_key_env_name: str | None = None,
        auth_token_env_name: str | None = None,
        endpoint_env_name: str | httpx.URL | None = None,
        aws_secret_key_env_name: str | None = None,
        aws_access_key_env_name: str | None = None,
        aws_region_env_name: str | None = None,
        aws_session_token_env_name: str | None = None,
        gc_region_env_name: str | NotGiven = NOT_GIVEN,
        gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
        gc_access_token_env_name: str | None = None,
        credentials: Any | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = 600,
        max_tokens: int = 2048,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,
        transport: Transport | None = None,
        proxies: ProxiesTypes | None = None,
        connection_pool_limits: httpx.Limits | None = None,
        _strict_response_validation: bool = False,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_k: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
    ):
        self.model_name = model_name
        self.api_type = api_type
        self.api_key_env_name = api_key_env_name
        self.auth_token_env_name = auth_token_env_name
        self.endpoint_env_name = endpoint_env_name
        self.aws_secret_key_env_name = aws_secret_key_env_name
        self.aws_access_key_env_name = aws_access_key_env_name
        self.aws_region_env_name = aws_region_env_name
        self.aws_session_token_env_name = aws_session_token_env_name
        self.timeout = timeout
        self.max_tokens = max_tokens
        self.max_retries = max_retries
        self.default_headers = default_headers
        self.default_query = default_query
        self.http_client = http_client
        self.transport = transport
        self.proxies = proxies
        self.connection_pool_limits = connection_pool_limits
        self._strict_response_validation = _strict_response_validation
        self.conversation_length_adjuster = conversation_length_adjuster
        self.system_instruction = system_instruction
        self.gc_region_env_name = gc_region_env_name
        self.gc_project_id_env_name = gc_project_id_env_name
        self.gc_access_token_env_name = gc_access_token_env_name
        self.credentials = credentials
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p

    def run(self, messages: list[dict[str, Any]]) -> CompletionResults:
        client = get_client(
            api_type=self.api_type,
            api_key_env_name=self.api_key_env_name,
            auth_token_env_name=self.auth_token_env_name,
            endpoint_env_name=self.endpoint_env_name,
            aws_secret_key_env_name=self.aws_secret_key_env_name,
            aws_access_key_env_name=self.aws_access_key_env_name,
            aws_region_env_name=self.aws_region_env_name,
            aws_session_token_env_name=self.aws_session_token_env_name,
            gc_region_env_name=self.gc_region_env_name,
            gc_project_id_env_name=self.gc_project_id_env_name,
            gc_access_token_env_name=self.gc_access_token_env_name,
            credentials=self.credentials,
            timeout=self.timeout,
            max_retries=self.max_retries,
            default_headers=self.default_headers,
            default_query=self.default_query,
            http_client=self.http_client,
            transport=self.transport,
            proxies=self.proxies,
            connection_pool_limits=self.connection_pool_limits,
            _strict_response_validation=self._strict_response_validation,
        )

        response = completion(
            client=client,
            model_name=self.model_name,
            messages=messages,
            system_instruction=self.system_instruction,
            max_tokens=self.max_tokens,
            timeout=self.timeout,
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
        )

        return CompletionResults(
            message=MessageParam(role=response.role, content=response.content),
            usage=Usage(
                model_name=response.model,
                prompt_tokens=response.usage.input_tokens,
                completion_tokens=response.usage.output_tokens,
            ),
            prompt=copy.deepcopy(messages),
        )

    async def arun(self, messages: list[dict[str, Any]]) -> CompletionResults:
        client = get_async_client(
            api_type=self.api_type,
            api_key_env_name=self.api_key_env_name,
            auth_token_env_name=self.auth_token_env_name,
            endpoint_env_name=self.endpoint_env_name,
            aws_secret_key_env_name=self.aws_secret_key_env_name,
            aws_access_key_env_name=self.aws_access_key_env_name,
            aws_region_env_name=self.aws_region_env_name,
            aws_session_token_env_name=self.aws_session_token_env_name,
            gc_region_env_name=self.gc_region_env_name,
            gc_project_id_env_name=self.gc_project_id_env_name,
            gc_access_token_env_name=self.gc_access_token_env_name,
            credentials=self.credentials,
            timeout=self.timeout,
            max_retries=self.max_retries,
            default_headers=self.default_headers,
            default_query=self.default_query,
            http_client=self.http_client,
            transport=self.transport,
            proxies=self.proxies,
            connection_pool_limits=self.connection_pool_limits,
            _strict_response_validation=self._strict_response_validation,
        )

        response = await acompletion(
            client=client,
            model_name=self.model_name,
            messages=messages,
            system_instruction=self.system_instruction,
            max_tokens=self.max_tokens,
            timeout=self.timeout,
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
        )

        return CompletionResults(
            message=MessageParam(role=response.role, content=response.content),
            usage=Usage(
                model_name=response.model,
                prompt_tokens=response.usage.input_tokens,
                completion_tokens=response.usage.output_tokens,
            ),
            prompt=copy.deepcopy(messages),
        )

    def stream(self, messages: list[dict[str, Any]]) -> Generator[CompletionResults, None, None]:
        client = get_client(
            api_type=self.api_type,
            api_key_env_name=self.api_key_env_name,
            auth_token_env_name=self.auth_token_env_name,
            endpoint_env_name=self.endpoint_env_name,
            aws_secret_key_env_name=self.aws_secret_key_env_name,
            aws_access_key_env_name=self.aws_access_key_env_name,
            aws_region_env_name=self.aws_region_env_name,
            aws_session_token_env_name=self.aws_session_token_env_name,
            gc_region_env_name=self.gc_region_env_name,
            gc_project_id_env_name=self.gc_project_id_env_name,
            gc_access_token_env_name=self.gc_access_token_env_name,
            credentials=self.credentials,
            timeout=self.timeout,
            max_retries=self.max_retries,
            default_headers=self.default_headers,
            default_query=self.default_query,
            http_client=self.http_client,
            transport=self.transport,
            proxies=self.proxies,
            connection_pool_limits=self.connection_pool_limits,
            _strict_response_validation=self._strict_response_validation,
        )

        response = completion(
            client=client,
            model_name=self.model_name,
            messages=messages,
            system_instruction=self.system_instruction,
            max_tokens=self.max_tokens,
            timeout=self.timeout,
            stream=True,
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
        )

        all_chunks = ""
        with response as stream:
            for r in stream:
                if isinstance(r, RawMessageStartEvent):
                    usage = Usage(
                        model_name=r.message.model,
                        prompt_tokens=r.message.usage.input_tokens,
                        completion_tokens=r.message.usage.output_tokens,
                    )
                    role = r.message.role
                elif isinstance(r, RawContentBlockStartEvent):
                    all_chunks += r.content_block.text
                elif isinstance(r, RawContentBlockDeltaEvent):
                    all_chunks += r.delta.text
                    yield CompletionResults(
                        message=MessageParam(
                            role=role, content=[TextBlockParam(text=all_chunks, type="text")]
                        ),
                        usage=Usage(),
                        prompt=[{}],
                    )
                elif isinstance(r, RawMessageStopEvent):
                    pass
                elif isinstance(r, RawMessageDeltaEvent):
                    usage += Usage(
                        completion_tokens=r.usage.output_tokens,
                    )

        yield CompletionResults(
            message=MessageParam(role=role, content=[TextBlockParam(text=all_chunks, type="text")]),
            usage=usage,
            prompt=copy.deepcopy(messages),
        )

    async def astream(
        self, messages: list[dict[str, Any]]
    ) -> AsyncGenerator[CompletionResults, None]:
        client = get_async_client(
            api_type=self.api_type,
            api_key_env_name=self.api_key_env_name,
            auth_token_env_name=self.auth_token_env_name,
            endpoint_env_name=self.endpoint_env_name,
            aws_secret_key_env_name=self.aws_secret_key_env_name,
            aws_access_key_env_name=self.aws_access_key_env_name,
            aws_region_env_name=self.aws_region_env_name,
            aws_session_token_env_name=self.aws_session_token_env_name,
            gc_region_env_name=self.gc_region_env_name,
            gc_project_id_env_name=self.gc_project_id_env_name,
            gc_access_token_env_name=self.gc_access_token_env_name,
            credentials=self.credentials,
            timeout=self.timeout,
            max_retries=self.max_retries,
            default_headers=self.default_headers,
            default_query=self.default_query,
            http_client=self.http_client,
            transport=self.transport,
            proxies=self.proxies,
            connection_pool_limits=self.connection_pool_limits,
            _strict_response_validation=self._strict_response_validation,
        )

        response = await acompletion(
            client=client,
            model_name=self.model_name,
            messages=messages,
            system_instruction=self.system_instruction,
            max_tokens=self.max_tokens,
            timeout=self.timeout,
            stream=True,
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
        )

        all_chunks = ""
        async with response as stream:
            async for r in stream:
                if isinstance(r, RawMessageStartEvent):
                    usage = Usage(
                        model_name=r.message.model,
                        prompt_tokens=r.message.usage.input_tokens,
                        completion_tokens=r.message.usage.output_tokens,
                    )
                    role = r.message.role
                elif isinstance(r, RawContentBlockStartEvent):
                    all_chunks += r.content_block.text
                elif isinstance(r, RawContentBlockDeltaEvent):
                    all_chunks += r.delta.text
                    yield CompletionResults(
                        message=MessageParam(
                            role=role, content=[TextBlockParam(text=all_chunks, type="text")]
                        ),
                        usage=Usage(),
                        prompt=[{}],
                    )
                elif isinstance(r, RawMessageStopEvent):
                    pass
                elif isinstance(r, RawMessageDeltaEvent):
                    usage += Usage(
                        completion_tokens=r.usage.output_tokens,
                    )

            yield CompletionResults(
                message=MessageParam(
                    role=role, content=[TextBlockParam(text=all_chunks, type="text")]
                ),
                usage=usage,
                prompt=copy.deepcopy(messages),
            )


class AnthropicChatModule(ChatWrapperModule):
    def __init__(
        self,
        model_name: str,
        api_type: str = "anthropic",
        api_key_env_name: str | None = None,
        auth_token_env_name: str | None = None,
        endpoint_env_name: str | httpx.URL | None = None,
        aws_secret_key_env_name: str | None = None,
        aws_access_key_env_name: str | None = None,
        aws_region_env_name: str | None = None,
        aws_session_token_env_name: str | None = None,
        gc_region_env_name: str | NotGiven = NOT_GIVEN,
        gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
        gc_access_token_env_name: str | None = None,
        credentials: Any | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = 600,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,
        transport: Transport | None = None,
        proxies: ProxiesTypes | None = None,
        connection_pool_limits: httpx.Limits | None = None,
        _strict_response_validation: bool = False,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        token_counter: TokenCounter | None = None,
        temperature: float | NotGiven = NOT_GIVEN,
        top_k: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
    ):
        # The module to call client API
        chat_model = AnthropicChatCoreModule(
            model_name=model_name,
            api_type=api_type,
            api_key_env_name=api_key_env_name,
            auth_token_env_name=auth_token_env_name,
            endpoint_env_name=endpoint_env_name,
            aws_secret_key_env_name=aws_secret_key_env_name,
            aws_access_key_env_name=aws_access_key_env_name,
            aws_region_env_name=aws_region_env_name,
            aws_session_token_env_name=aws_session_token_env_name,
            gc_region_env_name=gc_region_env_name,
            gc_project_id_env_name=gc_project_id_env_name,
            gc_access_token_env_name=gc_access_token_env_name,
            credentials=credentials,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
            conversation_length_adjuster=conversation_length_adjuster,
            system_instruction=system_instruction,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )

        super().__init__(
            chat_model=chat_model,
            conversation_memory=conversation_memory,
            content_filter=content_filter,
            token_counter=token_counter,
        )

    def _get_client_message_type(self) -> ClaudeMessage:
        return ClaudeMessage



from typing import Any, Iterable, Mapping, Union

import httpx
from anthropic import Anthropic, AnthropicBedrock, AsyncAnthropic, AsyncAnthropicBedrock
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import (
    NOT_GIVEN,
    Body,
    Headers,
    NotGiven,
    ProxiesTypes,
    Query,
    Timeout,
    Transport,
)
from anthropic.types import message_create_params
from anthropic.types.message_param import MessageParam
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.tool_param import ToolParam
from typing_extensions import Literal


def get_client(
    api_type: str,
    api_key_env_name: str | None = None,
    auth_token_env_name: str | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    aws_secret_key_env_name: str | None = None,
    aws_access_key_env_name: str | None = None,
    aws_region_env_name: str | None = None,
    aws_session_token_env_name: str | None = None,
    gc_region_env_name: str | NotGiven = NOT_GIVEN,
    gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
    gc_access_token_env_name: str | None = None,
    credentials: Any | None = None,
    timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    transport: Transport | None = None,
    proxies: ProxiesTypes | None = None,
    connection_pool_limits: httpx.Limits | None = None,
    _strict_response_validation: bool = False,
):
    if api_type == "anthropic":
        from .anthropic.client import get_anthropic_client

        return get_anthropic_client(
            api_key_env_name=api_key_env_name,
            auth_token_env_name=auth_token_env_name,
            endpoint_env_name=endpoint_env_name,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
        )
    elif api_type == "bedrock":
        from .bedrock.client import get_bedrock_client

        return get_bedrock_client(
            aws_secret_key_env_name=aws_secret_key_env_name,
            aws_access_key_env_name=aws_access_key_env_name,
            aws_region_env_name=aws_region_env_name,
            aws_session_token_env_name=aws_session_token_env_name,
            endpoint_env_name=endpoint_env_name,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            _strict_response_validation=_strict_response_validation,
        )
    elif api_type == "vertexai":
        from .vertexai.client import get_vertexai_client

        return get_vertexai_client(
            gc_region_env_name=gc_region_env_name,
            gc_project_id_env_name=gc_project_id_env_name,
            gc_access_token_env_name=gc_access_token_env_name,
            credentials=credentials,
            endpoint_env_name=endpoint_env_name,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
        )

    else:
        raise ValueError(f"Unknown API type: {api_type}")


def get_async_client(
    api_type: str,
    api_key_env_name: str | None = None,
    auth_token_env_name: str | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    aws_secret_key_env_name: str | None = None,
    aws_access_key_env_name: str | None = None,
    aws_region_env_name: str | None = None,
    aws_session_token_env_name: str | None = None,
    gc_region_env_name: str | NotGiven = NOT_GIVEN,
    gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
    gc_access_token_env_name: str | None = None,
    credentials: Any | None = None,
    timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    transport: Transport | None = None,
    proxies: ProxiesTypes | None = None,
    connection_pool_limits: httpx.Limits | None = None,
    _strict_response_validation: bool = False,
):
    if api_type == "anthropic":
        from .anthropic.client import get_async_anthropic_client

        return get_async_anthropic_client(
            api_key_env_name=api_key_env_name,
            auth_token_env_name=auth_token_env_name,
            endpoint_env_name=endpoint_env_name,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
        )
    elif api_type == "bedrock":
        from .bedrock.client import get_async_bedrock_client

        return get_async_bedrock_client(
            aws_secret_key_env_name=aws_secret_key_env_name,
            aws_access_key_env_name=aws_access_key_env_name,
            aws_region_env_name=aws_region_env_name,
            aws_session_token_env_name=aws_session_token_env_name,
            endpoint_env_name=endpoint_env_name,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            _strict_response_validation=_strict_response_validation,
        )
    elif api_type == "vertexai":
        from .vertexai.client import get_async_vertexai_client

        return get_async_vertexai_client(
            gc_region_env_name=gc_region_env_name,
            gc_project_id_env_name=gc_project_id_env_name,
            gc_access_token_env_name=gc_access_token_env_name,
            credentials=credentials,
            endpoint_env_name=endpoint_env_name,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
        )
    else:
        raise ValueError(f"Unknown API type: {api_type}")


def completion(
    client: Anthropic | AnthropicBedrock,
    model_name: Union[
        str,
        Literal[
            "claude-3-5-sonnet-20240620",
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
            "claude-2.1",
            "claude-2.0",
            "claude-instant-1.2",
        ],
    ],
    messages: Iterable[MessageParam],
    max_tokens: int = 2048,
    metadata: message_create_params.Metadata | NotGiven = NOT_GIVEN,
    stop_sequences: list[str] | NotGiven = NOT_GIVEN,
    stream: Literal[False] | Literal[True] | NotGiven = NOT_GIVEN,
    system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
    tool_choice: message_create_params.ToolChoice | NotGiven = NOT_GIVEN,
    tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
    extra_headers: Headers | None = None,
    extra_query: Query | None = None,
    extra_body: Body | None = None,
    timeout: float | httpx.Timeout | None | NotGiven = 600,
    temperature: float | NotGiven = NOT_GIVEN,
    top_k: int | NotGiven = NOT_GIVEN,
    top_p: float | NotGiven = NOT_GIVEN,
):
    return client.messages.create(
        model=model_name,
        messages=messages,
        max_tokens=max_tokens,
        metadata=metadata,
        stop_sequences=stop_sequences,
        stream=stream,
        system=system_instruction,
        temperature=temperature,
        tool_choice=tool_choice,
        tools=tools,
        top_p=top_p,
        top_k=top_k,
        extra_headers=extra_headers,
        extra_query=extra_query,
        extra_body=extra_body,
        timeout=timeout,
    )


async def acompletion(
    client: AsyncAnthropic | AsyncAnthropicBedrock,
    model_name: Union[
        str,
        Literal[
            "claude-3-5-sonnet-20240620",
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
            "claude-2.1",
            "claude-2.0",
            "claude-instant-1.2",
        ],
    ],
    messages: Iterable[MessageParam],
    max_tokens: int = 2048,
    metadata: message_create_params.Metadata | NotGiven = NOT_GIVEN,
    stop_sequences: list[str] | NotGiven = NOT_GIVEN,
    stream: Literal[False] | Literal[True] | NotGiven = NOT_GIVEN,
    system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
    tool_choice: message_create_params.ToolChoice | NotGiven = NOT_GIVEN,
    tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
    extra_headers: Headers | None = None,
    extra_query: Query | None = None,
    extra_body: Body | None = None,
    timeout: float | httpx.Timeout | None | NotGiven = 600,
    temperature: float | NotGiven = NOT_GIVEN,
    top_k: int | NotGiven = NOT_GIVEN,
    top_p: float | NotGiven = NOT_GIVEN,
):
    return await client.messages.create(
        model=model_name,
        messages=messages,
        max_tokens=max_tokens,
        metadata=metadata,
        stop_sequences=stop_sequences,
        stream=stream,
        system=system_instruction,
        temperature=temperature,
        tool_choice=tool_choice,
        tools=tools,
        top_p=top_p,
        top_k=top_k,
        extra_headers=extra_headers,
        extra_query=extra_query,
        extra_body=extra_body,
        timeout=timeout,
    )



import io
from pathlib import Path
from typing import Any, Literal

import numpy as np
from PIL import Image
from pydantic import BaseModel, ConfigDict, field_serializer, model_validator

from .types import PathType
from .utils import encode_image


class TextContent(BaseModel):
    text: str


class ImageContent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    image: Image.Image | PathType
    resolution: Literal["auto", "low", "high"] | None = None

    @property
    def format(self):
        if isinstance(self.image, Image.Image):
            return self.image.format.lower()
        elif isinstance(self.image, PathType):
            return Path(self.image).suffix.lstrip(".").lower()
        else:
            raise ValueError("Invalid image type")

    @model_validator(mode="before")
    def check_image_type(cls, data):
        if isinstance(data, dict) and "image" in data:
            if isinstance(data["image"], Image.Image):
                data["image"] = encode_image(data["image"], as_utf8=True)
            elif isinstance(data["image"], bytes):
                data["image"] = data["image"].decode("utf-8")
            elif isinstance(data["image"], str):
                try:
                    assert Path(data["image"]).is_file()
                    data["image"] = encode_image(Image.open(data["image"]), as_utf8=True)
                except Exception:
                    pass

        return data


class ToolCall(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str
    args: Any
    call_id: str | None = None


class ToolContent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    output: Any
    call_id: str | None = None
    args: Any | None = None
    funcname: str | None = None


class PDFContent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    file: PathType
    resolution_scale: float = 2.5
    image_resolution: Literal["auto", "low", "high"] = "high"

    def as_image_content(self) -> list[ImageContent]:
        from .file_utils.pdf import read_pdf_asimage

        return [
            ImageContent(image=image, resolution=self.image_resolution)
            for image in read_pdf_asimage(self.file, scale=self.resolution_scale)
        ]


class AudioContent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    data: PathType | bytes | np.ndarray
    sr: int | None = None
    mime_type: str | None = None

    def as_bytes(self) -> bytes:
        import soundfile as sf

        buffer = io.BytesIO()
        sf.write(
            buffer, self.data, samplerate=self.sr, format=self.mime_type.split("/")[-1].upper()
        )

        buffer.seek(0)
        return buffer.getvalue()

    @model_validator(mode="before")
    def setup(cls, data):
        try:
            import soundfile as sf
        except ModuleNotFoundError:
            return data

        if isinstance(data, dict) and "data" in data:
            try:
                assert Path(data["data"]).is_file()
                file_format = Path(data["data"]).suffix.lstrip(".").lower()
                if file_format in ["wav", "mp3", "aiff", "ogg", "flac"]:
                    data["mime_type"] = f"audio/{file_format}"
                elif file_format in ["mp4", "mpeg", "mov", "avi", "wmv", "mpg"]:
                    data["data"], data["sr"] = cls._from_video(data["data"])
                    data["mime_type"] = "audio/wav"
                else:
                    raise ValueError(f"Invalid audio file format: {file_format}")
            except Exception:
                pass

            if isinstance(data["data"], bytes):
                data["data"], data["sr"] = sf.read(io.BytesIO(data["data"]))
            elif isinstance(data["data"], io.BytesIO):
                data["data"], data["sr"] = sf.read(data["data"])
            elif isinstance(data["data"], (str | Path)):
                assert Path(data["data"]).is_file()
                data["data"], data["sr"] = sf.read(data["data"])
            elif isinstance(data["data"], np.ndarray):
                pass
            elif isinstance(data["data"], list):
                data["data"] = np.array(data["data"])
            else:
                raise ValueError("Invalid audio data type")

            data["data"] = cls._convert_stereo_to_mono(data["data"])

        return data

    @field_serializer("data")
    def serialize_data(self, data: np.ndarray) -> list:
        return data.tolist()

    @staticmethod
    def _convert_stereo_to_mono(data: np.ndarray) -> np.ndarray:
        if data.ndim == 2:
            return np.mean(data, axis=1)
        return data

    @staticmethod
    def _from_video(video_file: PathType) -> tuple[np.ndarray, int]:
        from moviepy.editor import VideoFileClip

        clip = VideoFileClip(Path(video_file).as_posix())
        audio = clip.audio
        fps = audio.fps
        audio_array = np.array(list(audio.iter_frames()))
        return audio_array, fps


class VideoContent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    file: PathType
    fps: float = 1.0
    image_resolution: Literal["auto", "low", "high"] = "low"
    include_audio: bool = False

    def as_image_content(self) -> list[ImageContent]:
        from .file_utils.video import sample_frames

        frames = [
            ImageContent(image=image, resolution=self.image_resolution)
            for image in sample_frames(self.file, fps=self.fps)
        ]

        if self.include_audio:
            frames.append(AudioContent(data=self.file))

        return frames


class URIContent(BaseModel):
    uri: str
    mime_type: str | None = None

    @model_validator(mode="before")
    def check_mime_type(cls, data):
        if isinstance(data, dict) and "uri" in data:
            file_format = Path(data["uri"]).suffix.lstrip(".").lower()
            if file_format in ["jpg", "jpeg", "png", "heic", "heif"]:
                if file_format == "jpg":
                    file_format = "jpeg"
                data["mime_type"] = f"image/{file_format}"
            elif file_format in ["pdf", "json"]:
                data["mime_type"] = f"application/{file_format}"
            elif file_format in ["wav", "mp3", "aiff", "aac", "ogg", "flac"]:
                data["mime_type"] = f"audio/{file_format}"
            elif file_format in ["mp4", "mpeg", "mov", "avi", "wmv", "mpg"]:
                data["mime_type"] = f"video/{file_format}"
            elif file_format in ["txt", "html", "csv", "markdown", "xml"]:
                data["mime_type"] = f"text/{file_format}"
        return data


ContentType = (
    str
    | TextContent
    | PDFContent
    | ToolContent
    | ImageContent
    | ToolCall
    | URIContent
    | VideoContent
    | AudioContent
)


class Message(BaseModel):
    role: str | None = None
    content: ContentType | list[ContentType]
    name: str | None = None


InputType = Message | list[Message] | ContentType | list[ContentType]
ConversationType = Message | list[Message]



from typing import Any

import numpy as np


class ConversationMixin:
    def load_conversation(self) -> list[dict[str, str]]:
        if self.conversation_memory is not None:
            messages: list[dict[str, str]] = self.conversation_memory.load()
        else:
            messages = []

        return messages

    def save_conversation(self, messages: list[dict[str, str]]) -> None:
        self.conversation_memory.store(messages)

    def _init_conversation_memory(self, init_conversation: list[dict[str, Any]] | None) -> None:
        if (
            self.conversation_memory
            and init_conversation
            and (
                not hasattr(self, "_INIT_STATUS")
                or (hasattr(self, "_INIT_STATUS") and not self._INIT_STATUS)
            )
        ):
            memory: list[dict[str, Any]] = self.conversation_memory.load()

            # check if init_conversation is already stored in conversation_memory
            # if not, add it to the conversation_memory
            if not np.isin(init_conversation, memory).all():
                memory.extend(init_conversation)
                self.conversation_memory.store(memory)
                self._INIT_STATUS = True


class FilterMixin:
    def apply_content_filter(self, messages: list[dict[str, str]]) -> list[dict[str, str]]:
        return self.content_filter.apply(messages)

    def restore_content_filter(self, messages: list[dict[str, str]]) -> list[dict[str, str]]:
        return self.content_filter.restore(messages)



import re
from typing import Any

from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionMessage,
    ChatCompletionMessageToolCall,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.chat_completion_content_part_image_param import ImageURL
from openai.types.chat.chat_completion_message_tool_call import Function

from ..base import BaseMessage
from ..message_content import ImageContent, Message, TextContent, ToolCall, ToolContent
from ..utils import decode_image, encode_image


class OpenAIMessage(BaseMessage):
    @property
    def as_user(self):
        return ChatCompletionUserMessageParam(
            role="user",
            content=self.contents,
            name=self.name if self.name else "User",
        )

    @property
    def as_assistant(self):
        return ChatCompletionAssistantMessageParam(
            role="assistant",
            content=self.contents,
            name=self.name if self.name else "Assistant",
        )

    @property
    def as_system(self):
        return ChatCompletionSystemMessageParam(
            role="system",
            content=self.contents,
            name=self.name if self.name else "System",
        )

    @property
    def as_function(self):
        return {
            "role": "tool",
            "tool_call_id": self.contents[0]["tool_call_id"],
            "name": self.contents[0]["name"],
            "content": self.contents[0]["content"],
        }

    @property
    def as_function_call(self) -> dict[str, Any]:
        return ChatCompletionMessage(
            role="assistant",
            tool_calls=[content for content in self.contents if content["type"] == "function"],
        ).model_dump()

    @staticmethod
    def _format_text_content(content: TextContent) -> ChatCompletionContentPartTextParam:
        return ChatCompletionContentPartTextParam(
            type="text",
            text=content.text,
        )

    @staticmethod
    def _format_image_content(content: ImageContent) -> ChatCompletionContentPartImageParam:
        file_format = decode_image(content.image, as_utf8=True).format.lower()
        return ChatCompletionContentPartImageParam(
            type="image_url",
            image_url=ImageURL(
                url=f"data:image/{file_format};base64,{content.image}",
                detail=content.resolution if content.resolution else "auto",
            ),
        )

    @staticmethod
    def _format_tool_content(content: ToolContent) -> dict[str, str]:
        return {
            "tool_call_id": "call_" + content.call_id.split("_")[-1],
            "name": content.funcname,
            "content": content.output,
        }

    @staticmethod
    def _format_tool_call_content(content: ToolCall) -> ChatCompletionMessageToolCall:
        return ChatCompletionMessageToolCall(
            id="call_" + content.call_id.split("_")[-1],
            type="function",
            function=Function(arguments=str(content.args), name=content.name),
        ).model_dump()

    @classmethod
    def from_client_message(cls, message: dict[str, Any]) -> Message:
        contents = message.get("content", [])
        common_contents = []
        for content in contents:
            if content["type"] == "text":
                common_contents.append(TextContent(text=content["text"]))
            elif content["type"] == "image_url":
                pattern = re.compile("^(data:image/.+;base64,)")
                url = pattern.sub("", content["image_url"]["url"])
                common_contents.append(
                    ImageContent(
                        image=decode_image(url, as_utf8=True),
                        resolution=content["image_url"]["detail"],
                    )
                )

            else:
                raise ValueError(f"Unknown content type: {content['type']}")

        return Message(
            role=message.get("role"),
            content=common_contents,
            name=message.get("name"),
        )

    @staticmethod
    def _preprocess_message(messages: list[Message]) -> list[Message]:
        new_messages = []
        for message in messages:
            role = message.role
            name = message.name
            content = message.content

            if role == "function":
                for content in message.content:
                    new_messages.append(
                        Message(
                            role=role,
                            content=[content],
                            name=name,
                        )
                    )
            else:
                new_messages.append(message)
        return new_messages



import datetime
import logging
import os
from pathlib import Path


class DefaultLogger:
    def __init__(
        self,
        path: str | Path | None = None,
        level: str = "DEBUG",
        include_timestamp: bool = False,
    ):
        if path is not None:
            path = Path(path)
            if path.exists():
                # recreate log file
                os.remove(path)

            path.parent.mkdir(parents=True, exist_ok=True)

            self.general_logger = logging.getLogger(path.as_posix())
        else:
            self.general_logger = logging.getLogger(__name__)

        # clear handlers
        if len(self.general_logger.handlers) > 0:
            self.general_logger.handlers.clear()

        stream_handler = logging.StreamHandler()
        self.general_logger.addHandler(stream_handler)

        if path is not None:
            file_general_handler = logging.FileHandler(path.as_posix())
            self.general_logger.addHandler(file_general_handler)

        self.general_logger.setLevel(getattr(logging, level))

        self.include_timestamp = include_timestamp

    def critical(self, message: str):
        if self.include_timestamp:
            message = self._add_timestamp(message)
            self.general_logger.critical(message)
        else:
            self.general_logger.critical(message)

    def fatal(self, message: str):
        if self.include_timestamp:
            message = self._add_timestamp(message)
            self.general_logger.fatal(message)
        else:
            self.general_logger.fatal(message)

    def error(self, message: str):
        if self.include_timestamp:
            message = self._add_timestamp(message)
            self.general_logger.error(message)
        else:
            self.general_logger.error(message)

    def warning(self, message: str):
        if self.include_timestamp:
            message = self._add_timestamp(message)
            self.general_logger.warning(message)
        else:
            self.general_logger.warning(message)

    def info(self, message: str):
        if self.include_timestamp:
            message = self._add_timestamp(message)
            self.general_logger.info(message)
        else:
            self.general_logger.info(message)

    def debug(self, message: str):
        if self.include_timestamp:
            message = self._add_timestamp(message)
            self.general_logger.debug(message)
        else:
            self.general_logger.debug(message)

    def _add_timestamp(self, message: str):
        return f"[{self.now_string}] {message}"

    @property
    def now_string(self):
        return str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))



from pydantic import field_validator

from ..tools import ToolConfig, ToolParameter, ToolProperty


class OpenAIToolProperty(ToolProperty):
    name: str
    type: str
    description: str
    enum: list[str | int | float] | None = None
    items: dict[str, str] | None = None

    def format(self):
        return {self.name: self.model_dump(exclude=["name"], exclude_none=True)}

    @field_validator("type")
    def check_type_value(cls, v):
        if v not in {"string", "number", "boolean", "array"}:
            raise ValueError("type must be one of string or number.")

        return v


class OpenAIToolParameter(ToolParameter):
    type: str = "object"
    properties: list[OpenAIToolProperty]
    required: list[str] | None = None

    def format(self):
        dumped = self.model_dump(exclude=["properties"], exclude_none=True)

        _properties = {}
        for p in self.properties:
            _properties.update(p.format())
        dumped["properties"] = _properties

        return dumped

    @field_validator("type")
    def check_type_value(cls, v):
        if v not in {"object"}:
            raise ValueError("supported type is only object")

        return v

    @field_validator("required")
    def check_required_value(cls, required, values):
        properties = values.data["properties"]
        property_names = {p.name for p in properties}
        if required is not None:
            for r in required:
                if r not in property_names:
                    raise ValueError(f"required property '{r}' is not defined in properties.")
        return required


class OpenAIToolConfig(ToolConfig):
    name: str
    type: str = "function"
    description: str
    parameters: OpenAIToolParameter
    strict: bool | None = None

    def format(self):
        dumped = self.model_dump(exclude=["parameters", "type", "strict"], exclude_none=True)
        dumped["parameters"] = self.parameters.format()

        output = {"type": self.type, self.type: dumped}

        if self.strict is not None:
            output["strict"] = self.strict

        return output

    @field_validator("type")
    def check_type_value(cls, v):
        if v not in {"function"}:
            raise ValueError("supported type is only function")

        return v

    @classmethod
    def from_universal_configs(cls, configs: list[ToolConfig]) -> list["OpenAIToolConfig"]:
        return [cls(**config.model_dump(exclude_none=True)) for config in configs]



from abc import ABC, abstractmethod
from inspect import isfunction, ismethod
from pathlib import Path
from typing import Any, AsyncGenerator, Callable, Generator

from PIL import Image
from pydantic import BaseModel

from .message_content import (
    AudioContent,
    ContentType,
    ImageContent,
    InputType,
    Message,
    PDFContent,
    TextContent,
    ToolCall,
    ToolContent,
    URIContent,
    VideoContent,
)
from .result import (
    CompletionResults,
    EmbeddingResults,
    FunctionCallingResults,
    ToolCallResponse,
    ToolOutput,
)
from .types import RoleType
from .utils import decode_image, is_valid_uri, model2func

ROLES = ["system", "user", "assistant", "function", "function_call", "tool"]
IMAGE_EXTENSIONS = ["jpg", "jpeg", "png", "heic", "heif"]
VIDEO_EXTENSIONS = ["mp4", "mpeg", "mov", "avi", "wmv", "mpg"]
AUDIO_EXTENSIONS = ["wav", "mp3", "aiff", "ogg", "flac"]


class BaseChatModule(ABC):
    @abstractmethod
    def run(self, messages: list[dict[str, str]]) -> CompletionResults:
        raise NotImplementedError

    async def arun(self, messages: list[dict[str, str]]) -> CompletionResults:
        raise NotImplementedError

    def stream(self, messages: list[dict[str, str]]) -> Generator[CompletionResults, None, None]:
        raise NotImplementedError

    async def astream(
        self, messages: list[dict[str, str]]
    ) -> AsyncGenerator[CompletionResults, None]:
        raise NotImplementedError


class BaseFunctionCallingModule(ABC):
    @abstractmethod
    def run(self, messages: list[dict[str, str]]) -> FunctionCallingResults:
        raise NotImplementedError

    async def arun(self, messages: list[dict[str, str]]) -> FunctionCallingResults:
        raise NotImplementedError

    def _set_runnable_tools_dict(self, tools: list[Callable | BaseModel]) -> dict[str, callable]:
        return {f.__name__: f if (isfunction(f) or ismethod(f)) else model2func(f) for f in tools}


class BaseEmbeddingModule(ABC):
    @abstractmethod
    def run(self, text: str | list[str]) -> EmbeddingResults:
        raise NotImplementedError

    async def arun(self, text: str | list[str]) -> EmbeddingResults:
        raise NotImplementedError


class BaseConversationLengthAdjuster(ABC):
    @abstractmethod
    def run(self, messages: list[dict[str, str]]) -> list[dict[str, str]]:
        raise NotImplementedError


class BaseFilter(ABC):
    @abstractmethod
    def apply(self, messages: list[dict[str, str]]) -> list[dict[str, str]]:
        raise NotImplementedError

    def aapply(self, messages: list[dict[str, str]]) -> list[dict[str, str]]:
        raise NotImplementedError

    @abstractmethod
    def restore(self, messages: list[dict[str, str]]) -> list[dict[str, str]]:
        raise NotImplementedError


class BaseConversationMemory(ABC):
    @abstractmethod
    def store(self, conversation_history: list[dict[str, str]]) -> None:
        raise NotImplementedError

    @abstractmethod
    def load(self, path: str) -> list[dict[str, str]]:
        raise NotImplementedError


class BaseMessage(ABC):
    def __init__(self, contents: list[dict[str, Any]], name: str | None = None):
        self.contents = contents
        self.name = name

    @property
    def as_system(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def as_user(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def as_assistant(self):
        raise NotImplementedError

    @property
    def as_function(self):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _format_text_content(content: TextContent) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_image_content(content: ImageContent) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_tool_content(content: ToolContent) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_tool_call_content(content: ToolCall) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_uri_content(content: str | Path) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_audio_content(content: str | Path) -> Any:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _from_completion_results(cls, results: CompletionResults) -> list[dict[str, str]]:
        raise NotImplementedError

    @classmethod
    def _from_completion_results(cls, response: CompletionResults) -> Message:
        return cls.from_client_message(response.message)

    @classmethod
    def _to_client_contents(
        cls,
        contents: ContentType | list[ContentType],
    ) -> list[dict[str, Any]]:
        if not isinstance(contents, list):
            contents = [contents]

        _contents = []

        for content in contents:
            if isinstance(content, TextContent):
                _contents.append(cls._format_text_content(content=content))
            elif isinstance(content, ImageContent):
                try:
                    _contents.append(cls._format_image_content(content=content))
                except NotImplementedError:
                    pass
            elif isinstance(content, ToolContent):
                try:
                    _contents.append(cls._format_tool_content(content=content))
                except NotImplementedError:
                    pass
            elif isinstance(content, ToolCall):
                try:
                    _contents.append(cls._format_tool_call_content(content=content))
                except NotImplementedError:
                    pass
            elif isinstance(content, (PDFContent, VideoContent)):
                try:
                    _contents.extend(
                        [
                            cls._format_image_content(content=image_content)
                            for image_content in content.as_image_content()
                        ]
                    )
                except NotImplementedError:
                    pass
            elif isinstance(content, URIContent):
                try:
                    _contents.append(cls._format_uri_content(content=content))
                except NotImplementedError:
                    pass
            elif isinstance(content, AudioContent):
                try:
                    _contents.append(cls._format_audio_content(content=content))
                except NotImplementedError:
                    pass
            else:
                raise NotImplementedError(f"Message type {type(content)} not implemented")

        return _contents

    @classmethod
    def to_client_message(
        cls,
        message: Message,
    ) -> Any:
        if message.role not in ROLES:
            raise ValueError(f"Invalid role: {message.role}")

        _contents = cls._to_client_contents(contents=message.content)

        return getattr(cls(contents=_contents, name=message.name), "as_" + message.role)

    @classmethod
    @abstractmethod
    def from_client_message(cls, message: Any) -> Message:
        raise NotImplementedError

    @classmethod
    def _string_to_universal_content(cls, content: str) -> ContentType:
        try:
            is_file = Path(content).is_file()
            is_uri = is_valid_uri(content)
            file_format = Path(content).suffix.lstrip(".").lower()
            if is_file:
                if file_format in IMAGE_EXTENSIONS:
                    return ImageContent(image=content)
                elif file_format in ["pdf"]:
                    return PDFContent(file=content)
                elif file_format in VIDEO_EXTENSIONS:
                    return VideoContent(file=content)
                elif file_format in AUDIO_EXTENSIONS:
                    return AudioContent(data=content)
                else:
                    raise ValueError(f"Unsupported file format: {file_format}")
            elif is_uri:
                return URIContent(uri=content)
            else:
                return TextContent(text=content)
        except OSError:
            try:
                decode_image(content, as_utf8=True)
                return ImageContent(image=content)
            except Exception:
                return TextContent(text=content)

    @classmethod
    def to_universal_contents(
        cls,
        contents: ContentType | list[ContentType],
    ) -> list[ContentType]:
        if not isinstance(contents, list):
            contents = [contents]

        _contents = []

        for content in contents:
            if isinstance(content, str):
                _string_content = cls._string_to_universal_content(content=content)
                if isinstance(_string_content, list):
                    _contents.extend(_string_content)
                else:
                    _contents.append(_string_content)
            elif isinstance(
                content,
                (
                    TextContent,
                    ImageContent,
                    ToolContent,
                    PDFContent,
                    ToolCall,
                    URIContent,
                    VideoContent,
                    AudioContent,
                ),
            ):
                _contents.append(content)
            else:
                raise NotImplementedError(f"Message type {type(content)} not implemented")

        return _contents

    @classmethod
    def to_universal_message(
        cls,
        message: InputType,
        role: RoleType | None = None,
        name: str | None = None,
    ) -> Message:
        if isinstance(message, Message):
            message.content = cls.to_universal_contents(contents=message.content)
            return message
        elif isinstance(message, ContentType):
            if role is None:
                raise ValueError("Role must be provided to create a message")

            message = Message(role=role, content=message, name=name)
            message.content = cls.to_universal_contents(contents=message.content)
            return message
        elif isinstance(message, list):
            if role is None:
                raise ValueError("Role must be provided to create a message")

            contents = []
            for m in message:
                _content = cls.to_universal_message(role=role, message=m, name=name).content
                contents.extend(_content)

            message = Message(role=role, content=contents, name=name)
            return message
        elif isinstance(message, Image.Image):
            return Message(
                role=role,
                content=[ImageContent(image=message)],
                name=name,
            )
        elif isinstance(message, dict):
            message = Message(**message)
            message.content = cls.to_universal_contents(contents=message.content)
            return message
        else:
            raise ValueError(f"Invalid message type {type(message)}")

    @classmethod
    def to_universal_message_from_completion_response(
        cls, response: CompletionResults
    ) -> dict[str, Any]:
        return cls._from_completion_results(response)

    @classmethod
    def to_universal_message_from_function_response(
        cls, response: FunctionCallingResults
    ) -> list[Message]:
        return [
            Message(
                role="function",  # global role
                content=[
                    ToolContent(
                        output=result.output,
                        args=result.args,
                        call_id=result.call_id,
                        funcname=result.funcname,
                    )
                ],
                name=result.funcname,
            )
            if isinstance(result, ToolOutput)
            else Message(
                role="assistant",  # global role
                content=[result],
            )
            for result in response.results
        ]

    @classmethod
    def to_universal_message_from_function_call(cls, response: FunctionCallingResults) -> Message:
        return Message(
            role="function_call",  # global role
            content=[
                ToolCall(
                    args=result.args,
                    name=result.name,
                    call_id=result.call_id,
                )
                if isinstance(result, ToolCallResponse)
                else result
                for result in response.calls
            ],
        )

    @staticmethod
    def _preprocess_message(messages: list[Message]) -> list[Message]:
        return messages


class BaseMetadataStore(ABC):
    @abstractmethod
    def store(self, ids: list[int], metadatas: list[dict[str, Any]]) -> None:
        raise NotImplementedError

    @abstractmethod
    def fetch(self, ids: list[int]) -> list[dict[str, Any]]:
        raise NotImplementedError


class BaseMetadataFilter(ABC):
    @abstractmethod
    def run(self, metadata: dict[str, str]) -> bool:
        raise NotImplementedError



from typing import Any, Sequence

from google.auth import credentials as auth_credentials


def get_model(
    model_name: str,
    max_output_tokens: int = 2048,
    json_mode: bool = False,
    api_key_env_name: str | None = None,
    api_type: str = "gemini",
    project_id_env_name: str | None = None,
    location_env_name: str | None = None,
    system_instruction: str | None = None,
    experiment: str | None = None,
    experiment_description: str | None = None,
    experiment_tensorboard: str | bool | None = None,
    staging_bucket: str | None = None,
    credentials: auth_credentials.Credentials | None = None,
    encryption_spec_key_name: str | None = None,
    network: str | None = None,
    service_account: str | None = None,
    endpoint_env_name: str | None = None,
    request_metadata: Sequence[tuple[str, str]] | None = None,
    response_schema: dict[str, Any] | None = None,
    presence_penalty: float | None = None,
    frequency_penalty: float | None = None,
    temperature: float | None = None,
    top_p: float | None = None,
    top_k: int | None = None,
    n_results: int = 1,
):
    if api_type == "genai":
        from .genai.client import get_genai_model

        return get_genai_model(
            model_name=model_name,
            api_key_env_name=api_key_env_name,
            system_instruction=system_instruction,
            max_output_tokens=max_output_tokens,
            json_mode=json_mode,
            response_schema=response_schema,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            n_results=n_results,
        )
    elif api_type == "vertexai":
        from .vertexai.client import get_vertexai_model

        return get_vertexai_model(
            model_name=model_name,
            max_output_tokens=max_output_tokens,
            system_instruction=system_instruction,
            json_mode=json_mode,
            project_id_env_name=project_id_env_name,
            location_env_name=location_env_name,
            experiment=experiment,
            experiment_description=experiment_description,
            experiment_tensorboard=experiment_tensorboard,
            staging_bucket=staging_bucket,
            credentials=credentials,
            encryption_spec_key_name=encryption_spec_key_name,
            network=network,
            service_account=service_account,
            endpoint_env_name=endpoint_env_name,
            request_metadata=request_metadata,
            response_schema=response_schema,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            n_results=n_results,
        )
    else:
        raise ValueError(f"Unknown API type: {api_type}")


def get_message_cls(api_type: str):
    if api_type == "genai":
        from .genai.message import GeminiMessage

        return GeminiMessage
    elif api_type == "vertexai":
        from .vertexai.message import VertexAIMessage

        return VertexAIMessage
    else:
        raise ValueError(f"Unknown API type: {api_type}")


def get_tool_cls(api_type: str):
    if api_type == "genai":
        from google.ai.generativelanguage import Tool

        return Tool

    elif api_type == "vertexai":
        from vertexai.generative_models import Tool

        return Tool
    else:
        raise ValueError(f"Unknown API type: {api_type}")


def get_client_tool_type(api_type: str):
    if api_type == "genai":
        from .genai.tools import GeminiToolConfig

        return GeminiToolConfig
    elif api_type == "vertexai":
        from .vertexai.tools import VertexAIToolConfig

        return VertexAIToolConfig
    else:
        raise ValueError(f"Unknown API type: {api_type}")


def get_call_config(api_type: str, tool_choice: str | None = "auto"):
    if api_type == "genai":
        from google.ai.generativelanguage import FunctionCallingConfig, ToolConfig

        if tool_choice is None:
            return ToolConfig(
                function_calling_config=FunctionCallingConfig(mode=FunctionCallingConfig.Mode.NONE)
            )
        elif tool_choice == "auto":
            return ToolConfig(
                function_calling_config=FunctionCallingConfig(mode=FunctionCallingConfig.Mode.AUTO)
            )
        else:
            return ToolConfig(
                function_calling_config=FunctionCallingConfig(
                    mode=FunctionCallingConfig.Mode.ANY,
                    allowed_function_names=[tool_choice],
                )
            )
    elif api_type == "vertexai":
        from vertexai.preview.generative_models import ToolConfig

        if tool_choice is None:
            return ToolConfig(
                function_calling_config=ToolConfig.FunctionCallingConfig(
                    mode=ToolConfig.FunctionCallingConfig.Mode.NONE
                )
            )
        elif tool_choice == "auto":
            return ToolConfig(
                function_calling_config=ToolConfig.FunctionCallingConfig(
                    mode=ToolConfig.FunctionCallingConfig.Mode.AUTO
                )
            )
        else:
            return ToolConfig(
                function_calling_config=ToolConfig.FunctionCallingConfig(
                    mode=ToolConfig.FunctionCallingConfig.Mode.ANY,
                    allowed_function_names=[tool_choice],
                )
            )
    else:
        raise ValueError(f"Unknown API type: {api_type}")



from azure.cosmos import exceptions, CosmosClient, PartitionKey
import logging
import os
import hashlib

from ..base import BaseConversationMemory

class CosmosConversationMemory(BaseConversationMemory):
    def __init__(self, endpoint_env_name: str, key_env_name: str, db_env_name: str, container_env_name: str):
        self.endpoint = os.getenv(endpoint_env_name)
        self.key = os.getenv(key_env_name)
        self.dbname = os.getenv(db_env_name)
        self.containername = os.getenv(container_env_name)
        # Create a Cosmos client
        try:
            client = CosmosClient(url=self.endpoint, credential=self.key)
        except:
            logging.error('Could not connect to Cosmos DB')
            raise ConnectionError
        # Get a database
        try:
            database = client.get_database_client(database=self.dbname)
        except exceptions.CosmosResourceNotFoundError:
            logging.error(f'Could not find database: {self.dbname}')
        # Get a container
        try:
            self.container = database.get_container_client(self.containername)
        except exceptions.CosmosResourceExistsError:
            logging.error(f'Could not find container: {self.containername}')

    def store(self, conversation_history: list[dict[str, str]]):
        for item in conversation_history:
            item["id"] = hashlib.sha256(str(item).encode()).hexdigest()
            try:
                self.container.create_item(item)
            except exceptions.CosmosResourceExistsError:
                pass
    
    def load(self) -> list[dict[str, str]]:
        result = []
        try:
            items = self.container.read_all_items()
            for item in items:
                result.append({k: v for k, v in item.items() if k in {"role", "content"}})
            return result
        except:
            return result



import copy

from ..base import BaseConversationMemory


class InMemoryConversationMemory(BaseConversationMemory):
    def __init__(self):
        self.history = []

    def store(self, conversation_history: list[dict[str, str]]):
        self.history = copy.deepcopy(conversation_history)

    def load(self) -> list[dict[str, str]]:
        return self.history



import base64
import json
from typing import Any

from google.ai.generativelanguage import (
    Blob,
    Content,
    FileData,
    FunctionCall,
    FunctionResponse,
    Part,
)

from ...base import BaseMessage
from ...message_content import (
    AudioContent,
    ImageContent,
    Message,
    TextContent,
    ToolCall,
    ToolContent,
    URIContent,
)
from ...utils import decode_image


class GeminiMessage(BaseMessage):
    @property
    def as_user(self) -> Content:
        return Content(role="user", parts=self.contents)

    @property
    def as_assistant(self) -> Content:
        return Content(role="model", parts=self.contents)

    @property
    def as_function(self) -> Content:
        return Content(
            role="function",
            parts=self.contents,
        )

    @property
    def as_function_call(self) -> Content:
        return Content(role="model", parts=self.contents)

    @staticmethod
    def _format_text_content(content: TextContent) -> Part:
        return Part(text=content.text)

    @staticmethod
    def _format_image_content(content: ImageContent) -> Part:
        file_format = decode_image(content.image, as_utf8=True).format.lower()
        _image_bytes = base64.b64decode(content.image.encode("utf-8"))

        image_bytes = Blob(mime_type=f"image/{file_format}", data=_image_bytes)
        return Part(inline_data=image_bytes)

    @staticmethod
    def _format_uri_content(content: URIContent) -> Part:
        file_data = FileData(file_uri=content.uri, mime_type=content.mime_type)
        return Part(file_data=file_data)

    @staticmethod
    def _format_audio_content(content: AudioContent) -> Part:
        _audio_bytes = content.as_bytes()
        audio_blob = Blob(data=_audio_bytes, mime_type=content.mime_type)
        return Part(inline_data=audio_blob)

    @staticmethod
    def _format_tool_content(content: ToolContent) -> Part:
        return Part(
            function_response=FunctionResponse(
                name=content.funcname, response={"content": content.output}
            )
        )

    @staticmethod
    def _format_tool_call_content(content: ToolCall) -> Part:
        return Part(
            function_call=FunctionCall(
                name=content.name,
                args=content.args if isinstance(content.args, dict) else json.loads(content.args),
            )
        )

    @classmethod
    def from_client_message(cls, message: Content) -> Message:
        serializable = cls._to_dict(message)

        common_contents = []

        for part in serializable.get("parts", []):
            if part.get("text"):
                common_contents.append(TextContent(text=part.get("text")))
            elif part.get("inlineData"):
                inline_data = part.get("inlineData", {})
                mime_type = inline_data.get("mimeType")
                file_format = mime_type.split("/")[1]
                if file_format in ["jpeg", "png", "jpg"]:
                    image_data = inline_data.get("data")
                    common_contents.append(
                        ImageContent(
                            image=image_data,
                        )
                    )
            else:
                raise ValueError(f"Unsupported part type: {part.get('type')}")

        return Message(
            role=serializable.get("role").replace("model", "assistant"),
            content=common_contents,
            name=serializable.get("name"),
        )

    @staticmethod
    def _to_dict(content: Content) -> dict[str, Any]:
        return json.loads(
            Content.to_json(
                content,
                including_default_value_fields=False,
                use_integers_for_enums=False,
            )
        )



import math
import os
import re
from typing import Any

import openai
import tiktoken
from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from ..utils import decode_image
from .model_config import (
    _TILE_SIZE,
    _TOKENS_PER_TILE,
    EMBEDDING_CONFIG,
    MODEL_CONFIG,
)

MODEL_ZOO = set(MODEL_CONFIG.keys()) | set(EMBEDDING_CONFIG.keys())


def get_encoding(model_name: str) -> tiktoken.Encoding:
    try:
        encoding = tiktoken.encoding_for_model(model_name)
    except KeyError:
        encoding = tiktoken.get_encoding("o200k_base")

    return encoding


def get_client(
    api_key_env_name: str,
    api_version: str | None = None,
    endpoint_env_name: str | None = None,
    organization_id_env_name: str | None = None,
    deployment_id_env_name: str | None = None,
    api_type: str | None = "openai",
    timeout: int = 60,
    max_retries: int = 5,
) -> OpenAI | AzureOpenAI:
    if api_type == "azure":
        assert (
            api_version and endpoint_env_name and deployment_id_env_name
        ), "api_version, endpoint_env_name, and deployment_id_env_name must be specified when api_type is 'azure'."
        return AzureOpenAI(
            **get_openai_client_settings(
                api_key_env_name=api_key_env_name,
                organization_id_env_name=organization_id_env_name,
                api_version=api_version,
                endpoint_env_name=endpoint_env_name,
                deployment_id_env_name=deployment_id_env_name,
                max_retries=max_retries,
                timeout=timeout,
            )
        )
    elif api_type == "openai":
        return OpenAI(
            **get_openai_client_settings(
                api_key_env_name=api_key_env_name,
                organization_id_env_name=organization_id_env_name,
                max_retries=max_retries,
                timeout=timeout,
            )
        )
    else:
        raise ValueError(f"api_type must be 'azure' or 'openai'. Got {api_type}.")


def get_async_client(
    api_key_env_name: str,
    api_version: str | None = None,
    endpoint_env_name: str | None = None,
    organization_id_env_name: str | None = None,
    deployment_id_env_name: str | None = None,
    api_type: str | None = "openai",
    timeout: int = 60,
    max_retries: int = 5,
) -> AsyncOpenAI | AsyncAzureOpenAI:
    if api_type == "azure":
        return AsyncAzureOpenAI(
            **get_openai_client_settings(
                api_key_env_name=api_key_env_name,
                organization_id_env_name=organization_id_env_name,
                api_version=api_version,
                endpoint_env_name=endpoint_env_name,
                deployment_id_env_name=deployment_id_env_name,
                max_retries=max_retries,
                timeout=timeout,
            )
        )
    elif api_type == "openai":
        return AsyncOpenAI(
            **get_openai_client_settings(
                api_key_env_name=api_key_env_name,
                organization_id_env_name=organization_id_env_name,
                max_retries=max_retries,
                timeout=timeout,
            )
        )
    else:
        raise ValueError(f"api_type must be 'azure' or 'openai'. Got {api_type}.")


def get_openai_client_settings(
    api_key_env_name: str,
    api_version: str | None = None,
    endpoint_env_name: str | None = None,
    organization_id_env_name: str | None = None,
    deployment_id_env_name: str | None = None,
    timeout: int = 60,
    max_retries: int = 5,
) -> dict[str, Any]:
    outputs = {}
    outputs["api_key"] = os.getenv(api_key_env_name)

    if isinstance(api_version, str):
        outputs["api_version"] = api_version

    if isinstance(endpoint_env_name, str):
        outputs["azure_endpoint"] = os.getenv(endpoint_env_name)

    if isinstance(organization_id_env_name, str):
        outputs["organization"] = os.getenv(organization_id_env_name)

    if isinstance(deployment_id_env_name, str):
        outputs["azure_deployment"] = os.getenv(deployment_id_env_name)

    outputs["timeout"] = timeout
    outputs["max_retries"] = max_retries

    return outputs


def set_openai_envs(
    api_key_env_name: str,
    api_version: str | None = None,
    api_type: str | None = None,
    endpoint_env_name: str | None = None,
    organization_id_env_name: str | None = None,
) -> None:
    openai.api_key = os.getenv(api_key_env_name)

    if isinstance(api_version, str):
        openai.api_version = api_version

    if isinstance(api_type, str):
        openai.api_type = api_type

    if isinstance(endpoint_env_name, str):
        openai.api_base = os.getenv(endpoint_env_name)

    if isinstance(organization_id_env_name, str):
        openai.organization = os.getenv(organization_id_env_name)


def get_n_tokens(
    message: dict[str, dict[str, str]], model_name: str, depth: int = 0
) -> dict[str, int]:
    """
    Return the number of tokens used by a list of messages.
    Forked and edited from : https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb

    FIXME: This function is not perfect. It might not be accurate while it works enough in many cases.
    """
    encoding = get_encoding(model_name)

    if model_name in MODEL_ZOO:
        if model_name == "gpt-3.5-turbo-0301":
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = -1  # if there's a name, the role is omitted
        else:
            tokens_per_message = 3
            tokens_per_name = 1
    else:
        raise NotImplementedError(
            f"get_n_tokens() is not implemented for model {model_name}. Please choose from following model : {', '.join(sorted(list(MODEL_ZOO)))}."
        )

    n_content_tokens = 0
    n_other_tokens = 0
    # num_tokens += tokens_per_message
    for key, value in message.items():
        if key == "tool_calls":
            continue
        if value is not None:
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        tokens = get_n_tokens(item, model_name, depth + 1)
                        n_content_tokens += tokens["content"]
                        n_other_tokens += tokens["other"]
                        # n_other_tokens += len(encoding.encode(key))
            elif key == "name":
                n_other_tokens += tokens_per_name
            elif isinstance(value, str):
                n_content_tokens += len(encoding.encode(value))
            elif key == "text":
                n_content_tokens += len(encoding.encode(item["text"]))
            elif key == "image_url":
                n_content_tokens += 85  # Base tokens
                if value["detail"] in ["high", "auto"]:
                    if value["url"].startswith("data:image/"):
                        img_encoded = re.sub("^(data:image/.+;base64,)", "", value["url"])
                        n_content_tokens += calculate_high_resolution_image_tokens(
                            decode_image(img_encoded).size
                        )
                    elif value["url"].startswith("https://"):
                        raise NotImplementedError(
                            "Image URL is not acceptable. Please use base64 encoded image."
                        )
            elif isinstance(value, dict):
                tokens = get_n_tokens(value, model_name, depth + 1)
                n_content_tokens += tokens["content"]
                n_other_tokens += tokens["other"]
                # n_other_tokens += len(encoding.encode(key))
            else:
                n_content_tokens += len(encoding.encode(value))
        else:
            n_other_tokens += len(encoding.encode(key))

    if depth == 0:
        n_other_tokens += tokens_per_message
        n_other_tokens += 3  # every reply is primed with <|start|>assistant<|message|>

    total_tokens = n_content_tokens + n_other_tokens
    return {"total": total_tokens, "content": n_content_tokens, "other": n_other_tokens}


def get_token_limit(model_name: str) -> int:
    if model_name in MODEL_ZOO:
        return MODEL_CONFIG[model_name]["max_tokens"]
    else:
        raise NotImplementedError(
            f"get_token_limit() is not implemented for model {model_name}. Please choose from following model : {', '.join(sorted(list(MODEL_ZOO)))}."
        )


def calculate_high_resolution_image_tokens(image_size: tuple[int, int] | list[int, int]) -> int:
    h, w = image_size
    short = min(h, w)
    long = max(h, w)

    if long > 2048:
        short = int(short * 2048 / long)
        long = 2048

    if short > 768:
        long = int(long * 768 / short)
        short = 768

    n_bins_long = math.ceil(long / _TILE_SIZE)
    n_bins_short = math.ceil(short / _TILE_SIZE)
    n_tiles = n_bins_long * n_bins_short
    return _TOKENS_PER_TILE * n_tiles



import copy
from typing import Any, AsyncGenerator, Generator, Optional, Sequence

from google.auth import credentials as auth_credentials

from ...base import (
    BaseChatModule,
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
    BaseMessage,
)
from ...llm_wrapper import ChatWrapperModule
from ...result import CompletionResults
from ...usage import TokenCounter, Usage
from ..gemini_utils import get_message_cls, get_model


class GeminiChatCoreModule(BaseChatModule):
    def __init__(
        self,
        model_name: str,
        api_key_env_name: str | None = None,
        api_type: str = "genai",
        project_id_env_name: str | None = None,
        location_env_name: str | None = None,
        experiment: str | None = None,
        experiment_description: str | None = None,
        experiment_tensorboard: str | bool | None = None,
        staging_bucket: str | None = None,
        credentials: auth_credentials.Credentials | None = None,
        encryption_spec_key_name: str | None = None,
        network: str | None = None,
        service_account: str | None = None,
        endpoint_env_name: str | None = None,
        request_metadata: Sequence[tuple[str, str]] | None = None,
        max_tokens: int = 2048,
        json_mode: bool = False,
        timeout: int = 60,
        system_instruction: str | None = None,
        response_schema: dict[str, Any] | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
    ):
        self.api_key_env_name = api_key_env_name
        self.model_name = model_name
        self.max_output_tokens = max_tokens
        self.api_type = api_type
        self.project_id_env_name = project_id_env_name
        self.location_env_name = location_env_name
        self.experiment = experiment
        self.experiment_description = experiment_description
        self.experiment_tensorboard = experiment_tensorboard
        self.staging_bucket = staging_bucket
        self.credentials = credentials
        self.encryption_spec_key_name = encryption_spec_key_name
        self.network = network
        self.service_account = service_account
        self.endpoint_env_name = endpoint_env_name
        self.request_metadata = request_metadata
        self.json_mode = json_mode
        self.response_schema = response_schema
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.temperature = temperature
        self.top_p = top_p
        self.top_k = top_k

        self.additional_kwargs = {}
        if api_type == "genai":
            from google.generativeai.types.helper_types import RequestOptions

            request_options = RequestOptions(
                timeout=timeout,
            )
            self.additional_kwargs["request_options"] = request_options

        self.system_instruction = system_instruction

    def run(
        self, messages: list[dict[str, str]], n_results: int | None = None
    ) -> CompletionResults:
        if n_results is not None and (self.api_type == "genai" and n_results > 1):
            raise ValueError("n_results > 1 is not supported for Google AI API")

        model = get_model(
            model_name=self.model_name,
            api_key_env_name=self.api_key_env_name,
            max_output_tokens=self.max_output_tokens,
            json_mode=self.json_mode,
            system_instruction=self.system_instruction,
            api_type=self.api_type,
            project_id_env_name=self.project_id_env_name,
            location_env_name=self.location_env_name,
            experiment=self.experiment,
            experiment_description=self.experiment_description,
            experiment_tensorboard=self.experiment_tensorboard,
            staging_bucket=self.staging_bucket,
            credentials=self.credentials,
            encryption_spec_key_name=self.encryption_spec_key_name,
            network=self.network,
            service_account=self.service_account,
            endpoint_env_name=self.endpoint_env_name,
            request_metadata=self.request_metadata,
            response_schema=self.response_schema,
            n_results=n_results,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )
        response = model.generate_content(contents=messages, **self.additional_kwargs)

        usage_metadata = response.usage_metadata
        parts = []
        candidates = response.candidates
        for candidate in candidates:
            content = candidate.content
            parts.extend(content.parts)

        if self.api_type == "genai":
            from google.ai.generativelanguage import Content

        else:
            from vertexai.generative_models import Content

        content = Content(role="model", parts=parts)

        return CompletionResults(
            message=content,
            usage=Usage(
                model_name=self.model_name,
                prompt_tokens=usage_metadata.prompt_token_count,
                completion_tokens=usage_metadata.candidates_token_count,
            ),
            prompt=copy.deepcopy(messages),
        )

    async def arun(
        self, messages: list[dict[str, str]], n_results: int | None = None
    ) -> CompletionResults:
        if n_results is not None and (self.api_type == "genai" and n_results > 1):
            raise ValueError("n_results > 1 is not supported for Google AI API")

        model = get_model(
            model_name=self.model_name,
            api_key_env_name=self.api_key_env_name,
            max_output_tokens=self.max_output_tokens,
            json_mode=self.json_mode,
            system_instruction=self.system_instruction,
            api_type=self.api_type,
            project_id_env_name=self.project_id_env_name,
            location_env_name=self.location_env_name,
            experiment=self.experiment,
            experiment_description=self.experiment_description,
            experiment_tensorboard=self.experiment_tensorboard,
            staging_bucket=self.staging_bucket,
            credentials=self.credentials,
            encryption_spec_key_name=self.encryption_spec_key_name,
            network=self.network,
            service_account=self.service_account,
            endpoint_env_name=self.endpoint_env_name,
            request_metadata=self.request_metadata,
            response_schema=self.response_schema,
            n_results=n_results,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )

        response = await model.generate_content_async(contents=messages, **self.additional_kwargs)

        usage_metadata = response.usage_metadata
        parts = []
        candidates = response.candidates
        for candidate in candidates:
            content = candidate.content
            parts.extend(content.parts)

        if self.api_type == "genai":
            from google.ai.generativelanguage import Content

        else:
            from vertexai.generative_models import Content

        content = Content(role="model", parts=parts)

        return CompletionResults(
            message=content,
            usage=Usage(
                model_name=self.model_name,
                prompt_tokens=usage_metadata.prompt_token_count,
                completion_tokens=usage_metadata.candidates_token_count,
            ),
            prompt=copy.deepcopy(messages),
        )

    def stream(
        self, messages: list[dict[str, str | list[str]]]
    ) -> Generator[CompletionResults, None, None]:
        model = get_model(
            model_name=self.model_name,
            api_key_env_name=self.api_key_env_name,
            max_output_tokens=self.max_output_tokens,
            json_mode=self.json_mode,
            system_instruction=self.system_instruction,
            api_type=self.api_type,
            project_id_env_name=self.project_id_env_name,
            location_env_name=self.location_env_name,
            experiment=self.experiment,
            experiment_description=self.experiment_description,
            experiment_tensorboard=self.experiment_tensorboard,
            staging_bucket=self.staging_bucket,
            credentials=self.credentials,
            encryption_spec_key_name=self.encryption_spec_key_name,
            network=self.network,
            service_account=self.service_account,
            endpoint_env_name=self.endpoint_env_name,
            request_metadata=self.request_metadata,
            response_schema=self.response_schema,
            n_results=1,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )
        responses = model.generate_content(contents=messages, stream=True, **self.additional_kwargs)

        chunk_all = ""
        for response in responses:
            content = response.candidates[0].content
            if content.parts[0].text:
                chunk_all += content.parts[0].text
                if hasattr(content.parts[0], "_raw_part"):
                    content.parts[0]._raw_part.text = chunk_all
                else:
                    content.parts[0].text = chunk_all

                last_content = content
                result = CompletionResults(
                    message=content,
                    usage=Usage(model_name=self.model_name),
                    prompt="",
                )

                yield result

        # at the end of the stream, return the entire response
        yield CompletionResults(
            message=last_content,
            usage=Usage(
                model_name=self.model_name,
                prompt_tokens=model.count_tokens(messages).total_tokens,
                completion_tokens=model.count_tokens(chunk_all).total_tokens,
            ),
            prompt=copy.deepcopy(messages),
        )

    async def astream(
        self, messages: list[dict[str, str]]
    ) -> AsyncGenerator[CompletionResults, None]:
        model = get_model(
            model_name=self.model_name,
            api_key_env_name=self.api_key_env_name,
            max_output_tokens=self.max_output_tokens,
            json_mode=self.json_mode,
            system_instruction=self.system_instruction,
            api_type=self.api_type,
            project_id_env_name=self.project_id_env_name,
            location_env_name=self.location_env_name,
            experiment=self.experiment,
            experiment_description=self.experiment_description,
            experiment_tensorboard=self.experiment_tensorboard,
            staging_bucket=self.staging_bucket,
            credentials=self.credentials,
            encryption_spec_key_name=self.encryption_spec_key_name,
            network=self.network,
            service_account=self.service_account,
            endpoint_env_name=self.endpoint_env_name,
            request_metadata=self.request_metadata,
            response_schema=self.response_schema,
            n_results=1,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )
        responses = await model.generate_content_async(
            contents=messages, stream=True, **self.additional_kwargs
        )

        chunk_all = ""
        async for _response in responses:
            content = _response.candidates[0].content
            if content.parts[0].text:
                chunk_all += content.parts[0].text
                if hasattr(content.parts[0], "_raw_part"):
                    content.parts[0]._raw_part.text = chunk_all
                else:
                    content.parts[0].text = chunk_all

                last_content = content

                result = CompletionResults(
                    message=content,
                    usage=Usage(model_name=self.model_name),
                    prompt="",
                )

                yield result

        # at the end of the stream, return the entire response
        yield CompletionResults(
            message=last_content,
            usage=Usage(
                model_name=self.model_name,
                prompt_tokens=(await model.count_tokens_async(messages)).total_tokens,
                completion_tokens=(await model.count_tokens_async(chunk_all)).total_tokens,
            ),
            prompt=copy.deepcopy(messages),
        )


class GeminiChatModule(ChatWrapperModule):
    def __init__(
        self,
        model_name: str,
        api_key_env_name: str | None = None,
        max_tokens: int = 2048,
        json_mode: bool = False,
        timeout: int = 60,
        content_filter: Optional[BaseFilter] = None,
        conversation_memory: Optional[BaseConversationMemory] = None,
        system_instruction: str | None = None,
        token_counter: TokenCounter | None = None,
        api_type: str = "genai",
        project_id_env_name: str | None = None,
        location_env_name: str | None = None,
        experiment: str | None = None,
        experiment_description: str | None = None,
        experiment_tensorboard: str | bool | None = None,
        staging_bucket: str | None = None,
        credentials: auth_credentials.Credentials | None = None,
        encryption_spec_key_name: str | None = None,
        network: str | None = None,
        service_account: str | None = None,
        endpoint_env_name: str | None = None,
        request_metadata: Sequence[tuple[str, str]] | None = None,
        response_schema: dict[str, Any] | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
    ):
        # The module to call client API
        chat_model = GeminiChatCoreModule(
            api_key_env_name=api_key_env_name,
            model_name=model_name,
            max_tokens=max_tokens,
            json_mode=json_mode,
            timeout=timeout,
            system_instruction=system_instruction,
            api_type=api_type,
            project_id_env_name=project_id_env_name,
            location_env_name=location_env_name,
            experiment=experiment,
            experiment_description=experiment_description,
            experiment_tensorboard=experiment_tensorboard,
            staging_bucket=staging_bucket,
            credentials=credentials,
            encryption_spec_key_name=encryption_spec_key_name,
            network=network,
            service_account=service_account,
            endpoint_env_name=endpoint_env_name,
            request_metadata=request_metadata,
            response_schema=response_schema,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
        )

        super().__init__(
            chat_model=chat_model,
            content_filter=content_filter,
            conversation_memory=conversation_memory,
            token_counter=token_counter,
        )

    def _get_client_message_type(self) -> type[BaseMessage]:
        return get_message_cls(self.chat_model.api_type)



import base64
import json
from typing import Any

from google.cloud.aiplatform_v1beta1.types import (
    content as gapic_content_types,
)
from google.cloud.aiplatform_v1beta1.types import tool as gapic_tool_types
from vertexai.generative_models import Content, Part

from ...base import BaseMessage
from ...message_content import (
    AudioContent,
    ImageContent,
    Message,
    TextContent,
    ToolCall,
    ToolContent,
    URIContent,
)
from ...utils import decode_image


class CustomPart(Part):
    @staticmethod
    def from_function_call(name: str, args: dict[str, Any]) -> "Part":
        return Part._from_gapic(
            raw_part=gapic_content_types.Part(
                function_call=gapic_tool_types.FunctionCall(
                    name=name,
                    args=args,
                )
            )
        )


class VertexAIMessage(BaseMessage):
    @property
    def as_user(self) -> Content:
        return Content(role="user", parts=self.contents)

    @property
    def as_assistant(self) -> Content:
        return Content(role="model", parts=self.contents)

    @property
    def as_function(self) -> Content:
        return Content(role="user", parts=self.contents)

    @property
    def as_function_call(self) -> Content:
        return Content(role="model", parts=self.contents)

    @staticmethod
    def _format_text_content(content: TextContent) -> Part:
        return Part.from_text(text=content.text)

    @staticmethod
    def _format_image_content(content: ImageContent) -> Part:
        file_format = decode_image(content.image, as_utf8=True).format.lower()
        _image_bytes = base64.b64decode(content.image.encode("utf-8"))

        return Part.from_data(mime_type=f"image/{file_format}", data=_image_bytes)

    @staticmethod
    def _format_uri_content(content: URIContent) -> Part:
        return Part.from_uri(uri=content.uri, mime_type=content.mime_type)

    @staticmethod
    def _format_audio_content(content: AudioContent) -> Part:
        _audio_bytes = content.as_bytes()
        return Part.from_data(mime_type=content.mime_type, data=_audio_bytes)

    @staticmethod
    def _format_tool_content(content: ToolContent) -> Part:
        return Part.from_function_response(
            name=content.funcname, response={"content": content.output}
        )

    @staticmethod
    def _format_tool_call_content(content: ToolCall) -> Part:
        return CustomPart.from_function_call(
            name=content.name,
            args=content.args if isinstance(content.args, dict) else json.loads(content.args),
        )

    @classmethod
    def from_client_message(cls, message: Content) -> Message:
        serializable = cls._to_dict(message)

        common_contents = []

        for part in serializable.get("parts", []):
            if part.get("text"):
                common_contents.append(TextContent(text=part.get("text")))
            elif part.get("inline_data"):
                inline_data = part.get("inline_data", {})
                mime_type = inline_data.get("mime_type")
                file_format = mime_type.split("/")[1]
                if file_format in ["jpeg", "png", "jpg"]:
                    image_data = inline_data.get("data")
                    common_contents.append(
                        ImageContent(
                            image=image_data,
                        )
                    )
            else:
                raise ValueError(f"Unsupported part type: {part.get('type')}")

        return Message(
            role=serializable.get("role").replace("model", "assistant"),
            content=common_contents,
            name=serializable.get("name"),
        )

    @staticmethod
    def _to_dict(content: Content) -> dict[str, Any]:
        return content.to_dict()



from os import PathLike
from pathlib import Path
from typing import Literal

from PIL import Image

PathType = Path | PathLike | str
ImageType = Image.Image | PathType
RoleType = Literal["system", "user", "assistant", "function", "function_call", "tool"]



_TOKENS_PER_TILE = 170
_TILE_SIZE = 512

_OLDER_MODEL_CONFIG = {
    "gpt-4-0613": {
        "max_tokens": 8192,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00003,
        "completion_cost_per_token": 0.00006,
    },
    "gpt-4-32k-0314": {
        "max_tokens": 32768,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00006,
        "completion_cost_per_token": 0.00012,
    },
    "gpt-4-32k-0613": {
        "max_tokens": 32768,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00006,
        "completion_cost_per_token": 0.00012,
    },
    "gpt-3.5-turbo-0301": {
        "max_tokens": 4096,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.0000015,
        "completion_cost_per_token": 0.000002,
    },
    "gpt-3.5-turbo-0613": {
        "max_tokens": 4096,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.0000015,
        "completion_cost_per_token": 0.000002,
    },
    "gpt-3.5-turbo-16k-0613": {
        "max_tokens": 16384,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.000003,
        "completion_cost_per_token": 0.000004,
    },
    "gpt-3.5-turbo-instruct": {
        "max_tokens": 4096,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.0000015,
        "completion_cost_per_token": 0.000002,
    },
}

_NEWER_MODEL_CONFIG = {
    "chatgpt-4o-latest": {
        "max_tokens": 128000,
        "max_output_tokens": 16384,
        "prompt_cost_per_token": 0.000005,
        "completion_cost_per_token": 0.000015,
    },
    "gpt-4o-2024-08-06": {
        "max_tokens": 128000,
        "max_output_tokens": 16384,
        "prompt_cost_per_token": 0.0000025,
        "completion_cost_per_token": 0.00001,
    },
    "gpt-4o-mini-2024-07-18": {
        "max_tokens": 128000,
        "max_output_tokens": 16384,
        "prompt_cost_per_token": 0.000000150,
        "completion_cost_per_token": 0.00000060,
    },
    "gpt-4o-2024-05-13": {
        "max_tokens": 128000,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.000005,
        "completion_cost_per_token": 0.000015,
    },
    "gpt-4-turbo-2024-04-09": {
        "max_tokens": 128000,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00001,
        "completion_cost_per_token": 0.00003,
    },
    "gpt-4-0125-preview": {
        "max_tokens": 128000,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00001,
        "completion_cost_per_token": 0.00003,
    },
    "gpt-4-1106-preview": {
        "max_tokens": 128000,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00001,
        "completion_cost_per_token": 0.00003,
    },
    "gpt-4-vision-preview": {
        "max_tokens": 128000,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.00001,
        "completion_cost_per_token": 0.00003,
    },
    "gpt-3.5-turbo-0125": {
        "max_tokens": 16384,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.0000005,
        "completion_cost_per_token": 0.0000015,
    },
    "gpt-3.5-turbo-1106": {
        "max_tokens": 16384,
        "max_output_tokens": 4096,
        "prompt_cost_per_token": 0.000001,
        "completion_cost_per_token": 0.000002,
    },
}

_NEWER_EMBEDDING_CONFIG = {
    "text-embedding-3-small": {
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.00000002,
    },
    "text-embedding-3-large": {
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.00000013,
    },
}

_OLDER_EMBEDDING_CONFIG = {
    "text-embedding-ada-002": {
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.0000001,
    },
}


EMBEDDING_CONFIG = {}
EMBEDDING_CONFIG.update(_OLDER_EMBEDDING_CONFIG)
EMBEDDING_CONFIG.update(_NEWER_EMBEDDING_CONFIG)

MODEL_CONFIG = {}
MODEL_CONFIG.update(_OLDER_MODEL_CONFIG)
MODEL_CONFIG.update(_NEWER_MODEL_CONFIG)

MODEL_POINT = {
    "gpt-4o-mini": "gpt-4o-mini-2024-07-18",
    "gpt-4o": "gpt-4o-2024-08-06",
    "gpt-4-turbo": "gpt-4-turbo-2024-04-09",
    "gpt-4": "gpt-4-0613",
    "gpt-4-32k": "gpt-4-32k-0613",
    "gpt-4-vision": "gpt-4-vision-preview",
    "gpt-3.5-turbo": "gpt-3.5-turbo-0125",
    "gpt-3.5-turbo-16k": "gpt-3.5-turbo-16k-0613",
}

_MODEL_POINT_CONFIG = {
    "gpt-4o-mini": MODEL_CONFIG[MODEL_POINT["gpt-4o-mini"]],
    "gpt-4o": MODEL_CONFIG[MODEL_POINT["gpt-4o"]],
    "gpt-4-turbo": MODEL_CONFIG[MODEL_POINT["gpt-4-turbo"]],
    "gpt-4": MODEL_CONFIG[MODEL_POINT["gpt-4"]],
    "gpt-4-32k": MODEL_CONFIG[MODEL_POINT["gpt-4-32k"]],
    "gpt-4-vision": MODEL_CONFIG[MODEL_POINT["gpt-4-vision"]],
    "gpt-3.5-turbo": MODEL_CONFIG[MODEL_POINT["gpt-3.5-turbo"]],
    "gpt-3.5-turbo-16k": MODEL_CONFIG[MODEL_POINT["gpt-3.5-turbo-16k"]],
}

MODEL_CONFIG.update(_MODEL_POINT_CONFIG)



import os
from typing import Mapping

import httpx
from anthropic import AnthropicBedrock, AsyncAnthropicBedrock
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import NOT_GIVEN, NotGiven


def get_bedrock_client(
    aws_secret_key_env_name: str | None = None,
    aws_access_key_env_name: str | None = None,
    aws_region_env_name: str | None = None,
    aws_session_token_env_name: str | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    _strict_response_validation: bool = False,
) -> AnthropicBedrock:
    return AnthropicBedrock(
        aws_secret_key=os.getenv(aws_secret_key_env_name) if aws_secret_key_env_name else None,
        aws_access_key=os.getenv(aws_access_key_env_name) if aws_access_key_env_name else None,
        aws_region=os.getenv(aws_region_env_name) if aws_region_env_name else None,
        aws_session_token=os.getenv(aws_session_token_env_name)
        if aws_session_token_env_name
        else None,
        base_url=os.getenv(endpoint_env_name) if endpoint_env_name else None,
        timeout=timeout,
        max_retries=max_retries,
        default_headers=default_headers,
        default_query=default_query,
        http_client=http_client,
        _strict_response_validation=_strict_response_validation,
    )


def get_async_bedrock_client(
    aws_secret_key_env_name: str | None = None,
    aws_access_key_env_name: str | None = None,
    aws_region_env_name: str | None = None,
    aws_session_token_env_name: str | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    _strict_response_validation: bool = False,
) -> AsyncAnthropicBedrock:
    return AsyncAnthropicBedrock(
        aws_secret_key=os.getenv(aws_secret_key_env_name) if aws_secret_key_env_name else None,
        aws_access_key=os.getenv(aws_access_key_env_name) if aws_access_key_env_name else None,
        aws_region=os.getenv(aws_region_env_name) if aws_region_env_name else None,
        aws_session_token=os.getenv(aws_session_token_env_name)
        if aws_session_token_env_name
        else None,
        base_url=os.getenv(endpoint_env_name) if endpoint_env_name else None,
        timeout=timeout,
        max_retries=max_retries,
        default_headers=default_headers,
        default_query=default_query,
        http_client=http_client,
        _strict_response_validation=_strict_response_validation,
    )



from collections import UserDict
from typing import Any

from pydantic import BaseModel, field_validator


class Usage(BaseModel):
    model_name: str = ""
    prompt_tokens: int = 0
    completion_tokens: int = 0

    def __add__(self, other: Any) -> "Usage":
        if isinstance(other, dict):
            other = Usage(**other)

        if (
            hasattr(other, "model_name")
            and other.model_name
            and self.model_name != other.model_name
        ):
            raise ValueError("model_name must be the same: {self.model_name} != {other.model_name}")

        if hasattr(other, "prompt_tokens"):
            prompt_tokens = self.prompt_tokens + other.prompt_tokens
        else:
            prompt_tokens = self.prompt_tokens
        if hasattr(other, "completion_tokens"):
            completion_tokens = self.completion_tokens + other.completion_tokens
        else:
            completion_tokens = self.completion_tokens
        return Usage(
            model_name=self.model_name,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
        )

    def __sub__(self, other: Any) -> "Usage":
        if isinstance(other, dict):
            other = Usage(**other)

        if (
            hasattr(other, "model_name")
            and other.model_name
            and self.model_name != other.model_name
        ):
            raise ValueError("model_name must be the same: {self.model_name} != {other.model_name}")

        if hasattr(other, "prompt_tokens"):
            prompt_tokens = self.prompt_tokens - other.prompt_tokens
        else:
            prompt_tokens = self.prompt_tokens
        if hasattr(other, "completion_tokens"):
            completion_tokens = self.completion_tokens - other.completion_tokens
        else:
            completion_tokens = self.completion_tokens
        return Usage(
            model_name=self.model_name,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
        )

    @property
    def total_tokens(self) -> int:
        return self.prompt_tokens + self.completion_tokens

    @field_validator("prompt_tokens")
    def check_prompt_tokens(cls, v):
        if v < 0:
            raise ValueError("prompt_tokens must be greater or equal to 0")
        return v

    @field_validator("completion_tokens")
    def check_completion_tokens(cls, v):
        if v < 0:
            raise ValueError("completion_tokens must be greater or equal to 0")
        return v

    def __repr__(self):
        return f"Usage(prompt_tokens={self.prompt_tokens}, completion_tokens={self.completion_tokens}, total_tokens={self.total_tokens})"


class TokenCounter(UserDict):
    def __init__(self, model_names: str | list[str] | None = None):
        tokens = {}
        if model_names is not None:
            if not isinstance(model_names, list):
                model_names = [model_names]

            for model_name in model_names:
                tokens[model_name] = Usage(model_name=model_name)

        super().__init__(tokens)

    def __add__(self, other: Usage):
        if other.model_name:
            if other.model_name not in self.data:
                self.data[other.model_name] = Usage(model_name=other.model_name)
            self.data[other.model_name] += other
        else:
            model_name = "unknown_model"
            if model_name not in self.data:
                self.data[model_name] = Usage(model_name=model_name)
            self.data[model_name] += other
        return self

    def __setitem__(self, model_name: str, usage: Usage):
        if model_name not in self.data:
            self.data[model_name] = Usage(model_name=model_name)
        self.data[model_name] += usage



from .assembly import ClaudeFunctionalChat
from .llm.chat import AnthropicChatModule
from .llm.function_calling import AnthropicFunctionCallingModule
from .message import ClaudeMessage
from .tools import ClaudeToolConfig, ClaudeToolParameter, ClaudeToolProperty



from anthropic.types import ToolParam
from pydantic import BaseModel

from ..tools import ToolConfig, ToolParameter, ToolProperty


class ClaudeToolProperty(ToolProperty):
    name: str
    type: str
    description: str
    enum: list[str | int | float] | None = None

    def format(self):
        return {self.name: self.model_dump(exclude=["name"], exclude_none=True)}


class ClaudeToolParameter(ToolParameter):
    type: str = "object"
    properties: list[ClaudeToolProperty]
    required: list[str | int | float] | None = None

    def format(self):
        dumped = self.model_dump(exclude=["properties"], exclude_none=True)

        _properties = {}
        for p in self.properties:
            _properties.update(p.format())
        dumped["properties"] = _properties

        return dumped


class ClaudeToolConfig(ToolConfig):
    name: str
    description: str
    parameters: ClaudeToolParameter

    def format(self):
        dumped = self.model_dump(exclude=["parameters"], exclude_none=True)
        dumped["input_schema"] = self.parameters.format()
        return ToolParam(**dumped)

    @classmethod
    def from_universal_configs(cls, configs: list[ToolConfig]) -> list["ClaudeToolConfig"]:
        return [cls(**config.model_dump(exclude_none=True)) for config in configs]



from copy import deepcopy
from typing import Any, AsyncGenerator, Generator

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI
from openai._types import NOT_GIVEN, NotGiven
from openai.types.chat import ChatCompletionAssistantMessageParam
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass

from ...base import (
    BaseChatModule,
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
)
from ...llm_wrapper import ChatWrapperModule
from ...result import CompletionResults
from ...usage import TokenCounter, Usage
from ..conversation_adjuster.truncate import OldConversationTruncationModule
from ..message import OpenAIMessage
from ..model_config import _OLDER_MODEL_CONFIG, MODEL_CONFIG, MODEL_POINT
from ..openai_utils import get_async_client, get_client, get_n_tokens, get_token_limit


def completion(
    client: OpenAI | AzureOpenAI,
    model_name: str,
    messages: Any,
    max_tokens: int,
    stream: bool,
    top_p: float | NotGiven = NOT_GIVEN,
    stop: str | NotGiven = NOT_GIVEN,
    frequency_penalty: float | NotGiven = NOT_GIVEN,
    n_results: int | NotGiven = NOT_GIVEN,
    presence_penalty: float | NotGiven = NOT_GIVEN,
    temperature: float | NotGiven = NOT_GIVEN,
    user: str | NotGiven = NOT_GIVEN,
    **kwargs,
):
    params = dict(
        model=model_name,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        stop=stop,
        stream=stream,
        n=n_results,
        user=user,
        **kwargs,
    )

    return client.chat.completions.create(**params)


async def acompletion(
    client: AsyncOpenAI | AsyncAzureOpenAI,
    model_name: str,
    messages: Any,
    max_tokens: int,
    stream: bool,
    top_p: float | NotGiven = NOT_GIVEN,
    stop: str | NotGiven = NOT_GIVEN,
    frequency_penalty: float | NotGiven = NOT_GIVEN,
    n_results: int | NotGiven = NOT_GIVEN,
    presence_penalty: float | NotGiven = NOT_GIVEN,
    temperature: float | NotGiven = NOT_GIVEN,
    user: str | NotGiven = NOT_GIVEN,
    **kwargs,
):
    params = dict(
        model=model_name,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        stop=stop,
        stream=stream,
        n=n_results,
        user=user,
        **kwargs,
    )

    return await client.chat.completions.create(**params)


def parse(
    client: OpenAI | AzureOpenAI,
    model_name: str,
    messages: Any,
    max_tokens: int,
    top_p: float | NotGiven = NOT_GIVEN,
    stop: str | NotGiven = NOT_GIVEN,
    frequency_penalty: float | NotGiven = NOT_GIVEN,
    n_results: int | NotGiven = NOT_GIVEN,
    presence_penalty: float | NotGiven = NOT_GIVEN,
    temperature: float | NotGiven = NOT_GIVEN,
    user: str | NotGiven = NOT_GIVEN,
    **kwargs,
):
    params = dict(
        model=model_name,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        stop=stop,
        n=n_results,
        user=user,
        **kwargs,
    )

    return client.beta.chat.completions.parse(**params)


async def aparse(
    client: AsyncOpenAI | AsyncAzureOpenAI,
    model_name: str,
    messages: Any,
    max_tokens: int,
    top_p: float | NotGiven = NOT_GIVEN,
    stop: str | NotGiven = NOT_GIVEN,
    frequency_penalty: float | NotGiven = NOT_GIVEN,
    n_results: int | NotGiven = NOT_GIVEN,
    presence_penalty: float | NotGiven = NOT_GIVEN,
    temperature: float | NotGiven = NOT_GIVEN,
    user: str | NotGiven = NOT_GIVEN,
    **kwargs,
):
    params = dict(
        model=model_name,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        stop=stop,
        n=n_results,
        user=user,
        **kwargs,
    )

    return await client.beta.chat.completions.parse(**params)


class OpenAIChatCoreModule(BaseChatModule):
    def __init__(
        self,
        api_key_env_name: str,
        model_name: str,
        organization_id_env_name: str | None = None,
        api_type: str = "openai",
        api_version: str | None = None,
        endpoint_env_name: str | None = None,
        deployment_id_env_name: str | None = None,
        max_tokens: int = 2048,
        timeout: int = 60,
        max_retries: int = 2,
        seed: int | NotGiven = NOT_GIVEN,
        json_mode: bool = False,
        system_instruction: str | None = None,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        top_p: float | NotGiven = NOT_GIVEN,
        frequency_penalty: float | NotGiven = NOT_GIVEN,
        presence_penalty: float | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        response_schema: BaseModel | None = None,
    ) -> None:
        assert api_type in ["openai", "azure"], "api_type must be 'openai' or 'azure'."
        if api_type == "azure":
            assert (
                api_version and endpoint_env_name and deployment_id_env_name
            ), "api_version, endpoint_env_name, and deployment_id_env_name must be specified for Azure API."

        self.api_key_env_name = api_key_env_name
        self.model_name = model_name
        self.organization_id_env_name = organization_id_env_name
        self.max_tokens = max_tokens
        self.timeout = timeout
        self.max_retries = max_retries
        self.api_type = api_type
        self.api_version = api_version
        self.endpoint_env_name = endpoint_env_name
        self.deployment_id_env_name = deployment_id_env_name
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.temperature = temperature
        self.user = user
        self.top_p = top_p

        self.additional_inputs = {}
        if model_name not in _OLDER_MODEL_CONFIG.keys():
            self.seed = seed
            self.additional_inputs["seed"] = seed
            if json_mode:
                if response_schema:
                    self.response_format = response_schema
                    self.additional_inputs["response_format"] = self.response_format
                else:
                    self.response_format = {"type": "json_object"} if json_mode else NOT_GIVEN
                    self.additional_inputs["response_format"] = self.response_format
            else:
                self.response_format = NOT_GIVEN
                self.additional_inputs["response_format"] = self.response_format

        else:
            # TODO : add logging message
            if seed:
                print(
                    f"seed is ignored because it's not supported for {model_name} (api_type:{api_type})"
                )
            if json_mode:
                print(
                    f"response_format is ignored because it's not supported for {model_name} (api_type:{api_type})"
                )

        if system_instruction:
            system_instruction = OpenAIMessage.to_universal_message(
                role="system", message=system_instruction
            )
            self.system_instruction = OpenAIMessage.to_client_message(system_instruction)
        else:
            self.system_instruction = None

        self.conversation_length_adjuster = conversation_length_adjuster

    def run(
        self, messages: list[dict[str, str]], n_results: int | NotGiven = NOT_GIVEN
    ) -> CompletionResults:
        if len(messages) == 0:
            raise ValueError("messages must not be empty.")

        if not isinstance(messages, list):
            raise ValueError("messages type must be list.")

        client = get_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        _messages = [self.system_instruction] + messages if self.system_instruction else messages
        if self.conversation_length_adjuster:
            _messages = self.conversation_length_adjuster.run(_messages)

        if isinstance(self.response_format, ModelMetaclass):
            response = parse(
                client=client,
                model_name=self.model_name,
                messages=_messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stop=NOT_GIVEN,
                n_results=n_results,
                user=self.user,
                **self.additional_inputs,
            )
        else:
            response = completion(
                client=client,
                model_name=self.model_name,
                messages=_messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stop=NOT_GIVEN,
                stream=False,
                n_results=n_results,
                user=self.user,
                **self.additional_inputs,
            )

        usage = Usage(model_name=self.model_name)
        usage += response.usage
        choices = response.choices
        contents = []
        for choice in choices:
            response_message = choice.message.content.strip("\n")
            contents.append({"type": "text", "text": response_message})

        return CompletionResults(
            usage=usage,
            message=ChatCompletionAssistantMessageParam(role="assistant", content=contents),
            prompt=deepcopy(_messages),
        )

    async def arun(
        self, messages: list[dict[str, str]], n_results: int | NotGiven = NOT_GIVEN
    ) -> CompletionResults:
        if len(messages) == 0:
            raise ValueError("messages must not be empty.")

        if not isinstance(messages, list):
            raise ValueError("messages type must be list.")

        client = get_async_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        _messages = [self.system_instruction] + messages if self.system_instruction else messages
        if self.conversation_length_adjuster:
            _messages = self.conversation_length_adjuster.run(_messages)

        if isinstance(self.response_format, ModelMetaclass):
            response = await aparse(
                client=client,
                model_name=self.model_name,
                messages=_messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stop=NOT_GIVEN,
                n_results=n_results,
                user=self.user,
                **self.additional_inputs,
            )
        else:
            response = await acompletion(
                client=client,
                model_name=self.model_name,
                messages=_messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stop=NOT_GIVEN,
                stream=False,
                n_results=n_results,
                user=self.user,
                **self.additional_inputs,
            )

        usage = Usage(model_name=self.model_name)
        usage += response.usage
        choices = response.choices
        contents = []
        for choice in choices:
            response_message = choice.message.content.strip("\n")
            contents.append({"type": "text", "text": response_message})

        return CompletionResults(
            usage=usage,
            message=ChatCompletionAssistantMessageParam(role="assistant", content=contents),
            prompt=deepcopy(_messages),
        )

    def stream(self, messages: list[dict[str, str]]) -> Generator[CompletionResults, None, None]:
        if len(messages) == 0:
            raise ValueError("messages must not be empty.")

        if not isinstance(messages, list):
            raise ValueError("messages type must be list.")

        client = get_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        _messages = [self.system_instruction] + messages if self.system_instruction else messages
        if self.conversation_length_adjuster:
            _messages = self.conversation_length_adjuster.run(_messages)

        if self.api_type == "openai":
            stream_options = {"include_usage": True}
            additional_inputs = self.additional_inputs | {"stream_options": stream_options}
        else:
            # Azure OpenAI does not support stream_options
            additional_inputs = self.additional_inputs

        response = completion(
            client=client,
            model_name=self.model_name,
            messages=_messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            stop=NOT_GIVEN,
            stream=True,
            n_results=1,
            user=self.user,
            **additional_inputs,
        )

        all_chunk = ""
        prompt_tokens = 0
        completion_tokens = 0
        for r in response:
            if len(r.choices) > 0:
                delta = r.choices[0].delta
                if delta is not None:
                    chunk = delta.content
                    if chunk is not None:
                        all_chunk += chunk
                        yield CompletionResults(
                            usage=Usage(model_name=self.model_name),
                            message=ChatCompletionAssistantMessageParam(
                                role="assistant",
                                content=[{"type": "text", "text": all_chunk}],
                            ),
                            prompt=[{}],
                        )
                    else:
                        if self.api_type == "azure":
                            # Azure OpenAI does not return prompt_tokens and completion_tokens
                            prompt_tokens = sum(
                                [get_n_tokens(m, self.model_name)["total"] for m in messages]
                            )

                            completion_tokens = get_n_tokens(
                                ChatCompletionAssistantMessageParam(
                                    role="assistant",
                                    content=[{"type": "text", "text": all_chunk}],
                                ),
                                self.model_name,
                            )["total"]

            else:
                if self.api_type == "openai":
                    prompt_tokens = r.usage.prompt_tokens
                    completion_tokens = r.usage.completion_tokens

        # at the end of stream, return the whole message and usage
        usage = Usage(
            model_name=self.model_name,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
        )

        yield CompletionResults(
            usage=usage,
            message=ChatCompletionAssistantMessageParam(
                role="assistant",
                content=[{"type": "text", "text": all_chunk}],
            ),
            prompt=deepcopy(_messages),
        )

    async def astream(
        self, messages: list[dict[str, str]]
    ) -> AsyncGenerator[CompletionResults, None]:
        if len(messages) == 0:
            raise ValueError("messages must not be empty.")

        if not isinstance(messages, list):
            raise ValueError("messages type must be list.")

        client = get_async_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        _messages = [self.system_instruction] + messages if self.system_instruction else messages
        if self.conversation_length_adjuster:
            _messages = self.conversation_length_adjuster.run(_messages)

        if self.api_type == "openai":
            stream_options = {"include_usage": True}
            additional_inputs = self.additional_inputs | {"stream_options": stream_options}
        else:
            # Azure OpenAI does not support stream_options
            additional_inputs = self.additional_inputs

        response = await acompletion(
            client=client,
            model_name=self.model_name,
            messages=_messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            stop=NOT_GIVEN,
            stream=True,
            n_results=1,
            user=self.user,
            **additional_inputs,
        )

        all_chunk = ""
        prompt_tokens = 0
        completion_tokens = 0
        async for r in response:
            if len(r.choices) > 0:
                delta = r.choices[0].delta
                if delta is not None:
                    chunk = delta.content
                    if chunk is not None:
                        all_chunk += chunk
                        yield CompletionResults(
                            usage=Usage(model_name=self.model_name),
                            message=ChatCompletionAssistantMessageParam(
                                role="assistant",
                                content=[{"type": "text", "text": all_chunk}],
                            ),
                            prompt=[{}],
                        )
                    else:
                        if self.api_type == "azure":
                            # Azure OpenAI does not return prompt_tokens and completion_tokens
                            prompt_tokens = sum(
                                [get_n_tokens(m, self.model_name)["total"] for m in messages]
                            )

                            completion_tokens = get_n_tokens(
                                ChatCompletionAssistantMessageParam(
                                    role="assistant",
                                    content=[{"type": "text", "text": all_chunk}],
                                ),
                                self.model_name,
                            )["total"]

            else:
                if self.api_type == "openai":
                    prompt_tokens = r.usage.prompt_tokens
                    completion_tokens = r.usage.completion_tokens

        # at the end of stream, return the whole message and usage
        usage = Usage(
            model_name=self.model_name,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
        )

        yield CompletionResults(
            usage=usage,
            message=ChatCompletionAssistantMessageParam(
                role="assistant",
                content=[{"type": "text", "text": all_chunk}],
            ),
            prompt=deepcopy(_messages),
        )


class OpenAIChatModule(ChatWrapperModule):
    def __init__(
        self,
        api_key_env_name: str,
        model_name: str,
        api_type: str = "openai",
        api_version: str | None = None,
        endpoint_env_name: str | None = None,
        deployment_id_env_name: str | None = None,
        organization_id_env_name: str | None = None,
        max_tokens: int | None = 2048,
        timeout: int = 60,
        max_retries: int = 2,
        seed: int | NotGiven = NOT_GIVEN,
        json_mode: bool = False,
        context_length: int | None = None,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        system_instruction: str | None = None,
        token_counter: TokenCounter | None = None,
        top_p: float | None | NotGiven = NOT_GIVEN,
        frequency_penalty: float | None | NotGiven = NOT_GIVEN,
        presence_penalty: float | None | NotGiven = NOT_GIVEN,
        temperature: float | None | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        response_schema: BaseModel | None = None,
    ):
        if model_name in MODEL_POINT.keys():
            print(f"{model_name} is automatically converted to {MODEL_POINT[model_name]}")
            model_name = MODEL_POINT[model_name]

        assert (
            model_name in MODEL_CONFIG.keys()
        ), f"model_name must be one of {', '.join(sorted(MODEL_CONFIG.keys()))}."

        token_lim = get_token_limit(model_name)
        max_tokens = max_tokens if max_tokens else MODEL_CONFIG[model_name]["max_output_tokens"]
        context_length = token_lim - max_tokens if context_length is None else context_length
        assert (
            token_lim >= max_tokens + context_length
        ), f"max_tokens({max_tokens}) + context_length({context_length}) must be less than or equal to the token limit of the model ({token_lim})."
        assert context_length > 0, "context_length must be positive."

        conversation_length_adjuster = (
            OldConversationTruncationModule(model_name=model_name, context_length=context_length)
            if conversation_length_adjuster is None
            else conversation_length_adjuster
        )

        # The module to call client API
        chat_model = OpenAIChatCoreModule(
            api_key_env_name=api_key_env_name,
            organization_id_env_name=organization_id_env_name,
            model_name=model_name,
            max_tokens=max_tokens,
            api_type=api_type,
            api_version=api_version,
            endpoint_env_name=endpoint_env_name,
            deployment_id_env_name=deployment_id_env_name,
            timeout=timeout,
            max_retries=max_retries,
            seed=seed,
            json_mode=json_mode,
            system_instruction=system_instruction,
            conversation_length_adjuster=conversation_length_adjuster,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            temperature=temperature,
            user=user,
            response_schema=response_schema,
        )

        super().__init__(
            chat_model=chat_model,
            conversation_memory=conversation_memory,
            content_filter=content_filter,
            token_counter=token_counter,
        )

    def _get_client_message_type(self) -> OpenAIMessage:
        return OpenAIMessage



from openai.resources.embeddings import Embeddings

from ..base import BaseEmbeddingModule
from ..result import EmbeddingResults
from ..usage import Usage
from ..utils import make_batch
from .model_config import _NEWER_EMBEDDING_CONFIG, EMBEDDING_CONFIG
from .openai_utils import get_async_client, get_client


class OpenAIEmbeddingModule(BaseEmbeddingModule):
    def __init__(
        self,
        api_key_env_name: str,
        organization_id_env_name: str | None = None,
        model_name: str = "text-embedding-ada-002",
        dimensions: int | None = None,
        user: str | None = None,
        api_type: str | None = "openai",
        api_version: str | None = None,
        endpoint_env_name: str | None = None,
        deployment_id_env_name: str | None = None,
        max_retries: int = 5,
        timeout: int = 60,
        batch_size: int = 10,
    ):
        assert api_type in ["openai", "azure"], "api_type must be 'openai' or 'azure'."
        if api_type == "azure":
            assert (
                api_version and endpoint_env_name and deployment_id_env_name
            ), "api_version, endpoint_env_name, and deployment_id_env_name must be specified for Azure API."

        assert (
            model_name in EMBEDDING_CONFIG.keys()
        ), f"model_name must be one of {', '.join(sorted(EMBEDDING_CONFIG.keys()))}."

        self.api_key_env_name = api_key_env_name
        self.organization_id_env_name = organization_id_env_name
        self.model_name = model_name
        self.api_type = api_type
        self.api_version = api_version
        self.endpoint_env_name = endpoint_env_name
        self.deployment_id_env_name = deployment_id_env_name
        self.max_retries = max_retries
        self.timeout = timeout
        self.batch_size = batch_size

        self.additional_params = {}
        if dimensions is not None:
            if model_name in _NEWER_EMBEDDING_CONFIG:
                self.additional_params["dimensions"] = dimensions
            else:
                print(f"Warning: dimensions is not supported for {model_name}. It will be ignored.")
        if user is not None:
            self.additional_params["user"] = user

    def run(self, text: str | list[str]) -> EmbeddingResults:
        client = get_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        embedder = Embeddings(client)

        if not isinstance(text, list):
            text = [text]

        embeddings = []
        total_usage = Usage(model_name=self.model_name)
        for batch in make_batch(text, batch_size=self.batch_size):
            response = embedder.create(input=batch, model=self.model_name, **self.additional_params)
            embeddings.extend([e.embedding for e in response.data])
            total_usage += response.usage

        results = EmbeddingResults(
            text=text,
            embeddings=embeddings,
            usage=total_usage,
        )
        return results

    async def arun(self, text: str) -> EmbeddingResults:
        client = get_async_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        embedder = Embeddings(client)

        if not isinstance(text, list):
            text = [text]

        embeddings = []
        total_usage = Usage(model_name=self.model_name)
        for batch in make_batch(text, batch_size=self.batch_size):
            response = await embedder.create(
                input=batch, model=self.model_name, **self.additional_params
            )
            embeddings.extend([e.embedding for e in response.data])
            total_usage += response.usage

        results = EmbeddingResults(
            text=text,
            embeddings=embeddings,
            usage=total_usage,
        )
        return results



import copy
import json
from typing import Any, Callable, Optional, Sequence

from google.auth import credentials as auth_credentials
from pydantic import BaseModel

from ...base import (
    BaseChatModule,
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
    BaseFunctionCallingModule,
    BaseMessage,
)
from ...llm_wrapper import FunctionCallingWrapperModule
from ...result import FunctionCallingResults, ToolCallResponse, ToolOutput
from ...tools import ToolConfig
from ...usage import TokenCounter, Usage
from ...utils import generate_dummy_call_id
from ..gemini_utils import (
    get_call_config,
    get_client_tool_type,
    get_message_cls,
    get_model,
    get_tool_cls,
)


class GeminiFunctionCallingCoreModule(BaseFunctionCallingModule):
    def __init__(
        self,
        model_name: str,
        tools: list[Callable],
        tool_configs: list[ToolConfig],
        api_key_env_name: str | None = None,
        api_type: str = "genai",
        project_id_env_name: str | None = None,
        location_env_name: str | None = None,
        experiment: str | None = None,
        experiment_description: str | None = None,
        experiment_tensorboard: str | bool | None = None,
        staging_bucket: str | None = None,
        credentials: auth_credentials.Credentials | None = None,
        encryption_spec_key_name: str | None = None,
        network: str | None = None,
        service_account: str | None = None,
        endpoint_env_name: str | None = None,
        request_metadata: Sequence[tuple[str, str]] | None = None,
        max_tokens: int = 2048,
        json_mode: bool = False,
        timeout: int = 60,
        system_instruction: str | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
    ):
        self.api_key_env_name = api_key_env_name
        self.model_name = model_name
        self.max_output_tokens = max_tokens
        self.api_type = api_type
        self.project_id_env_name = project_id_env_name
        self.location_env_name = location_env_name
        self.experiment = experiment
        self.experiment_description = experiment_description
        self.experiment_tensorboard = experiment_tensorboard
        self.staging_bucket = staging_bucket
        self.credentials = credentials
        self.encryption_spec_key_name = encryption_spec_key_name
        self.network = network
        self.service_account = service_account
        self.endpoint_env_name = endpoint_env_name
        self.request_metadata = request_metadata
        self.json_mode = json_mode
        self.system_instruction = system_instruction
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.temperature = temperature
        self.top_p = top_p
        self.top_k = top_k

        self.additional_kwargs = {}
        if api_type == "genai":
            from google.generativeai.types.helper_types import RequestOptions

            request_options = RequestOptions(
                timeout=timeout,
            )
            self.additional_kwargs["request_options"] = request_options

        ClientToolConfig = self._get_client_tool_config_type(api_type)
        client_tool_configs = ClientToolConfig.from_universal_configs(tool_configs)
        self.tools = self._set_runnable_tools_dict(tools)

        tool_cls = get_tool_cls(api_type=api_type)
        function_declarations = [config.format() for config in client_tool_configs]
        self.tool_configs = [tool_cls(function_declarations=function_declarations)]

    def _get_call_config(self, tool_choice: str | None = "auto"):
        return get_call_config(api_type=self.api_type, tool_choice=tool_choice)

    def _get_client_tool_config_type(self, api_type: str):
        return get_client_tool_type(api_type=api_type)

    def run(
        self, messages: list[dict[str, str]], tool_choice: list[str] | str | None = "auto"
    ) -> FunctionCallingResults:
        model = get_model(
            model_name=self.model_name,
            api_key_env_name=self.api_key_env_name,
            max_output_tokens=self.max_output_tokens,
            json_mode=self.json_mode,
            system_instruction=self.system_instruction,
            api_type=self.api_type,
            project_id_env_name=self.project_id_env_name,
            location_env_name=self.location_env_name,
            experiment=self.experiment,
            experiment_description=self.experiment_description,
            experiment_tensorboard=self.experiment_tensorboard,
            staging_bucket=self.staging_bucket,
            credentials=self.credentials,
            encryption_spec_key_name=self.encryption_spec_key_name,
            network=self.network,
            service_account=self.service_account,
            endpoint_env_name=self.endpoint_env_name,
            request_metadata=self.request_metadata,
            n_results=1,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )

        call_config = self._get_call_config(tool_choice=tool_choice)

        response = model.generate_content(
            contents=messages,
            tools=self.tool_configs,
            tool_config=call_config,
            **self.additional_kwargs,
        )
        parts = response.candidates[0].content.parts

        results = []
        calls = []
        for part in parts:
            if fn := part.function_call:
                funcname = fn.name
                args = dict(fn.args)
                func_out = self.tools[funcname](**args)
                dummy_call_id = generate_dummy_call_id(24)
                output = ToolOutput(
                    call_id=dummy_call_id,
                    funcname=funcname,
                    args=json.dumps(args),
                    output=func_out,
                )

                call = ToolCallResponse(
                    name=funcname,
                    args=args,
                    call_id=dummy_call_id,
                )
                results.append(output)
                calls.append(call)

        usage_metadata = response.usage_metadata

        return FunctionCallingResults(
            usage=Usage(
                model_name=self.model_name,
                prompt_tokens=usage_metadata.prompt_token_count,
                completion_tokens=usage_metadata.candidates_token_count,
            ),
            results=results,
            calls=calls,
            prompt=copy.deepcopy(messages),
        )

    async def arun(
        self, messages: list[dict[str, str]], tool_choice: list[str] | str | None = "auto"
    ) -> FunctionCallingResults:
        model = get_model(
            model_name=self.model_name,
            api_key_env_name=self.api_key_env_name,
            max_output_tokens=self.max_output_tokens,
            json_mode=self.json_mode,
            system_instruction=self.system_instruction,
            api_type=self.api_type,
            project_id_env_name=self.project_id_env_name,
            location_env_name=self.location_env_name,
            experiment=self.experiment,
            experiment_description=self.experiment_description,
            experiment_tensorboard=self.experiment_tensorboard,
            staging_bucket=self.staging_bucket,
            credentials=self.credentials,
            encryption_spec_key_name=self.encryption_spec_key_name,
            network=self.network,
            service_account=self.service_account,
            endpoint_env_name=self.endpoint_env_name,
            request_metadata=self.request_metadata,
            n_results=1,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )

        call_config = self._get_call_config(tool_choice=tool_choice)

        response = await model.generate_content_async(
            contents=messages,
            tools=self.tool_configs,
            tool_config=call_config,
            **self.additional_kwargs,
        )
        parts = response.candidates[0].content.parts

        results = []
        calls = []
        for part in parts:
            if fn := part.function_call:
                funcname = fn.name
                args = dict(fn.args)
                func_out = self.tools[funcname](**args)
                dummy_call_id = generate_dummy_call_id(24)
                output = ToolOutput(
                    call_id=dummy_call_id,
                    funcname=funcname,
                    args=json.dumps(args),
                    output=func_out,
                )

                call = ToolCallResponse(
                    name=funcname,
                    args=args,
                    call_id=dummy_call_id,
                )
                results.append(output)
                calls.append(call)

        usage_metadata = response.usage_metadata

        return FunctionCallingResults(
            usage=Usage(
                model_name=self.model_name,
                prompt_tokens=usage_metadata.prompt_token_count,
                completion_tokens=usage_metadata.candidates_token_count,
            ),
            results=results,
            calls=calls,
            prompt=copy.deepcopy(messages),
        )


class GeminiFunctionCallingModule(FunctionCallingWrapperModule):
    def __init__(
        self,
        model_name: str,
        tools: list[Callable],
        tool_configs: list[ToolConfig],
        api_key_env_name: str | None = None,
        max_tokens: int = 2048,
        json_mode: bool = False,
        timeout: int = 60,
        content_filter: Optional[BaseFilter] = None,
        conversation_memory: Optional[BaseConversationMemory] = None,
        token_counter: Optional[TokenCounter] = None,
        api_type: str = "genai",
        project_id_env_name: str | None = None,
        location_env_name: str | None = None,
        experiment: str | None = None,
        experiment_description: str | None = None,
        experiment_tensorboard: str | bool | None = None,
        staging_bucket: str | None = None,
        credentials: auth_credentials.Credentials | None = None,
        encryption_spec_key_name: str | None = None,
        network: str | None = None,
        service_account: str | None = None,
        endpoint_env_name: str | None = None,
        request_metadata: Sequence[tuple[str, str]] | None = None,
        system_instruction: str | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
    ):
        # The module to call client API
        function_calling_model = GeminiFunctionCallingCoreModule(
            api_key_env_name=api_key_env_name,
            model_name=model_name,
            max_tokens=max_tokens,
            json_mode=json_mode,
            timeout=timeout,
            tools=tools,
            tool_configs=tool_configs,
            api_type=api_type,
            project_id_env_name=project_id_env_name,
            location_env_name=location_env_name,
            experiment=experiment,
            experiment_description=experiment_description,
            experiment_tensorboard=experiment_tensorboard,
            staging_bucket=staging_bucket,
            credentials=credentials,
            encryption_spec_key_name=encryption_spec_key_name,
            network=network,
            service_account=service_account,
            endpoint_env_name=endpoint_env_name,
            request_metadata=request_metadata,
            system_instruction=system_instruction,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
        )

        super().__init__(
            function_calling_model=function_calling_model,
            content_filter=content_filter,
            conversation_memory=conversation_memory,
            token_counter=token_counter,
        )

    def _get_client_message_type(self) -> type[BaseMessage]:
        return get_message_cls(self.function_calling_model.api_type)



from typing import Any, Optional

from pydantic import BaseModel

from .usage import Usage


class CompletionResults(BaseModel):
    message: Any
    usage: Usage = Usage()
    prompt: Any = None


class ToolCallResponse(BaseModel):
    name: str
    args: Any
    call_id: str | None = None


class ToolOutput(BaseModel):
    call_id: str | None
    funcname: str | None
    args: Any | None
    output: Any


class FunctionCallingResults(BaseModel):
    usage: Usage
    results: list[Any]
    calls: Any | None = None
    prompt: Any | None = None


class EmbeddingResults(BaseModel):
    text: list[str]
    embeddings: list[list[float]]
    usage: Usage


class RetrievalResults(BaseModel):
    ids: list[int | str]
    documents: list[str]
    metadatas: Optional[list[dict[str, Any]] | list[None]]
    scores: list[float]
    collections: list[str]
    usage: Usage = Usage(prompt_tokens=0, completion_tokens=0)



from .assembly import OpenAIFunctionalChat
from .embedding import OpenAIEmbeddingModule
from .llm.chat import OpenAIChatModule
from .llm.function_calling import (
    OpenAIFunctionCallingModule,
    ToolOutput,
)
from .message import OpenAIMessage
from .openai_utils import get_n_tokens
from .tools import OpenAIToolConfig, OpenAIToolParameter, OpenAIToolProperty



from .base import (
    BaseChatModule,
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseEmbeddingModule,
    BaseFilter,
    BaseFunctionCallingModule,
    BaseMessage,
)
from .memory.in_memory import InMemoryConversationMemory
from .memory.json import JSONConversationMemory
from .message_content import (
    AudioContent,
    ContentType,
    ImageContent,
    Message,
    PDFContent,
    TextContent,
    ToolCall,
    ToolContent,
    URIContent,
    VideoContent,
)
from .prompt_template import PromptTemplate
from .result import CompletionResults, EmbeddingResults, FunctionCallingResults, RetrievalResults
from .tools import ToolConfig, ToolParameter, ToolProperty
from .usage import TokenCounter, Usage



from typing import Any, Optional

from pydantic import BaseModel


class ToolProperty(BaseModel):
    name: str
    type: str
    description: str
    enum: list[str | int | float] | None = None
    items: dict[str, Any] | None = None


class ToolParameter(BaseModel):
    type: str = "object"
    properties: list[ToolProperty]
    required: Optional[list[str]] = None


class ToolConfig(BaseModel):
    name: str
    description: str
    parameters: ToolParameter
    strict: bool | None = None

    @classmethod
    def from_pydantic(cls, model: BaseModel, **kwargs) -> "ToolConfig":
        configs = {}
        params = {}

        configs["name"] = model.__name__
        schema = model.model_json_schema()
        defs = schema.pop("$defs", {})
        params["required"] = schema.pop("required", None)
        params["type"] = schema.pop("type", "object")
        configs["description"] = schema.pop("description", "No description.")

        properties = schema.get("properties", {})
        props = []
        for k, v in properties.items():
            if ref := v.get("$ref"):
                ref = ref.split("/")[-1]
                v = defs[ref]

            props.append(
                {
                    "name": k,
                    "type": v.get("type"),
                    "description": v.get("description"),
                    "enum": v.get("enum"),
                }
            )
        params["properties"] = props
        configs["parameters"] = params

        return cls(**configs, **kwargs)
